import os
from os.path import isfile, join
import re
import numpy as np

def get_phoneme_label_vector(phn_file, length, fs, befaft, mode='phonemes', return_label_lists=False):
    '''
    Creates a time-series vector of phoneme labels based on the output of the Penn-Phonetic Alignment
    procedure.
    
    Returns label vector as numpy array of shape (time, ) containing categorical labels.
    Assumes that the .phn files were made on wav files that did not have a befaft zero period.

    This function can be used instead of the ``get_label_vecs_from_files`` method
    inside the ``Aligner`` class if you only want a single file's label vector.

    Parameters
    ----------
    phn_file : string, path to .phn file
    length : int, length of stimulus (in samples) including befaft periods
    fs : int, sampling rate
    befaft : array or list of length 2, containing befaft time periods in seconds
    mode : string, either 'phonemes', or 'manner'. Any empty or 'sp' periods are labeled -1.
        If 'manner': labels correspond to ['plosive','fricative','nasal','sonorant']
    return_label_lists : bool, if True, returns a 2-tuple containing the labels as well as the list
        of possible labels, so that the index of a label in the list is the integer label assigned to it.
        
    Returns
    -------
    labels : np.array, shape (time,)
        Integer labels over time. Full length is befaft[0]*fs+length+befaft[1]*fs
        
    item_list : list
        List of phonemes (if mode=='phonemes') or manners of articulation (if mode=='manner')
        which were used. The index of a phoneme indicates the label it is assigned.
    '''

    pattern = r'[0-9]'
    phoneme_list = ['EH','K','S','L','AH','M','EY','SH','N','P','OY','T','OW','Z','W','D','B','V','IH','AA','R','AY','ER','AE','AO','NG','G','TH','IY','F','DH','HH','UH','CH','UW','AW','JH','Y','ZH']
    manner_list = ['plosive','fricative','nasal','sonorant']
    manner_dict = {'EH':'sonorant',
                    'K':'plosive',
                    'S':'fricative',
                    'L':'sonorant',
                    'AH':'sonorant',
                    'M':'nasal',
                    'EY':'sonorant',
                    'SH':'fricative',
                    'N':'nasal',
                    'P':'plosive',
                    'OY':'sonorant',
                    'T':'plosive',
                    'OW':'sonorant',
                    'Z':'fricative',
                    'W':'sonorant',
                    'D':'plosive',
                    'B':'plosive',
                    'V':'fricative',
                    'IH':'sonorant',
                    'AA':'sonorant',
                    'R':'sonorant',
                    'AY':'sonorant',
                    'ER':'sonorant',
                    'AE':'sonorant',
                    'AO':'sonorant',
                    'NG':'nasal',
                    'G':'plosive',
                    'TH':'fricative',
                    'IY':'sonorant',
                    'F':'fricative',
                    'DH':'fricative',
                    'HH':'fricative',
                    'UH':'sonorant',
                    'CH':'velar',
                    'UW':'sonorant',
                    'AW':'sonorant',
                    'JH':'velar',
                    'Y':'sonorant',
                    'ZH':'sonorant',
                    'sp':'none'}
    
    f = open(phn_file, 'r')
    Lines = [line.strip() for line in f.readlines()]
    f.close()

    labels = -1 * np.ones((length,))

    for line in Lines:
        phn = re.sub(pattern, '', line.split(' ')[-1])
        start_time, end_time = line.split(' ')[:2]
        start_time, end_time = float(start_time), float(end_time)
        start_time += befaft[0]
        end_time += befaft[0]
        if mode == 'phonemes':
            labels[round(start_time*fs):round(end_time*fs)] = phoneme_list.index(phn) if phn in phoneme_list else -1.0
        elif mode == 'manner':
            labels[round(start_time*fs):round(end_time*fs)] = manner_list.index(manner_dict[phn]) if manner_dict[phn] in manner_list else -1.0
        else:
            raise Exception(f"mode parameter must be either 'phonemes' or 'manner', but found {mode}")
        

    if return_label_lists:
        if mode == 'phonemes':
            return labels, phoneme_list
        elif mode == 'manner':
            return labels, manner_list
        else:
            raise Exception(f"mode parameter must be either 'phonemes' or 'manner', but found {mode}")
    return labels

def get_word_label_vector(wrd_file, length, fs, befaft, wrd_dict=None, wrd_files_dir=None, return_wrd_dict=False):
    '''
    Returns label vector as numpy array of shape (time, ) containing categorical labels.
    Assumes that the .wrd files were made on wav files that did not have a befaft zero period.

    This function can be used instead of the ``get_label_vecs_from_files`` method
    inside the ``Aligner`` class if you only want a single file's label vector.

    Parameters
    ----------
    wrd_file : string, path to .wrd file
    length : int
        length of stimulus (in samples) including befaft periods
    fs : int, sampling rate
    befaft : array or list of length 2
        containing befaft time periods in seconds
    wrd_dict : dict
        keys are words (capitalized) and values are integers which become the labels for each word
    wrd_files_dir : string
        Path to directory containing all the .wrd files to be included in the dictionary. This is
        ignored if wrd_dict is supplied, but otherwise, a new wrd_dict is created using the .wrd
        files in this directory. By default, this will make 'sp' (space) have a label of -1.
    return_wrd_dict : bool
        Whether or not to return the wrd_dict along with the labels as a tuple
        
    Returns
    -------
    labels : np.array, shape (time,)
        Integer labels over time. Full length is befaft[0]*fs+length+befaft[1]*fs
    wrd_dict : dict
        Word dictionary (word to integer) used to create the labels.
        Only returned if return_wrd_dict==True
    '''
    
    pattern = r'[0-9]'
    
    if wrd_dict is None and wrd_files_dir is None:
        raise Exception('Must provide either wrd_dict or wrd_files_dir')
    elif wrd_dict is None:
        wrd_dict = create_wrd_dict(wrd_files_dir)
    
    f = open(wrd_file, 'r')
    Lines = [line.strip() for line in f.readlines()]
    f.close()

    labels = -1 * np.ones((length,))

    for line in Lines:
        wrd = re.sub(pattern, '', line.split(' ')[-1])
        start_time, end_time = line.split(' ')[:2]
        start_time, end_time = float(start_time), float(end_time)
        start_time += befaft[0]
        end_time += befaft[0]
        try:
            labels[round(start_time*fs):round(end_time*fs)] = wrd_dict[wrd]
        
        except KeyError:
            raise KeyError(f"{wrd} is not a valid key in wrd_dict")

    if return_wrd_dict:
        return labels, wrd_dict

    return labels

def create_wrd_dict(wrd_files_dir, list_to_skip=[]):
    '''
    Create a new word to label dictionary, which can be passed to get_word_label_vector.
    
    Parameters
    ----------
    wrd_files_dir : string
        Path to directory containing all the .wrd files to be used
    list_to_skip : list of strings, default=[]
        Words which will not be added to the dictionary.
        
    Returns
    -------
    wrd_dict : dict
        Dictionary of word:int (key:value) pairs for all the words in the corpus
        of files in the directory.
    '''
    
    pattern = r'[0-9]'
    
    wrd_files = sorted([f for f in os.listdir(wrd_files_dir) if isfile(join(wrd_files_dir, f)) and f.endswith('.wrd')])
    
    wrd_dict = {}
    
    for wrd_file in wrd_files:
        f = open(join(wrd_files_dir, wrd_file), 'r')
        Lines = [line.strip() for line in f.readlines()]
        f.close()
    
        for line in Lines:
            wrd = re.sub(pattern, '', line.split(' ')[-1])
            if wrd == 'sp':
                wrd_dict[wrd] = -1
            elif wrd not in wrd_dict and wrd not in list_to_skip:
                wrd_dict[wrd] = len(wrd_dict)

    return wrd_dict
    


import numpy as np
import scipy.cluster.hierarchy as shc
from sklearn.cluster import AgglomerativeClustering
import matplotlib.pyplot as plt

def shadederrorplot(x, y, ax=None, err_method='stderr', plt_args={}, shade_args={}, nan_policy='omit'):
    '''
    Parameters
    ----------
    x : shape (time,)
    y : shape (time, n_samples)
    ax : plt axes to use
    err_method : string, default='stderr
        One of ['stderr','std'], the method to use to calculate error bars.
    plt_args : dict
        Args to be passed to plt.plot(). e.g. 'color','linewidth',...
    shade_args : dict
        Args to be passed to plt.fill_between(). e.g. 'color','alpha',...
    nan_policy : string, default='omit'
        One of ['omit','raise','propogate']. If 'omit', will ignore any nan in the
        inputs, if 'raise', will raise a ValueError if nan is found in input, if
        'propogate', do not do anything special with nan values.
        
    Raises
    ------
    ValueError
        if nan found in input
    '''
    if nan_policy not in ['omit','raise','propogate']:
        raise Exception(f"nan_policy must be one of ['omit','raise','propogate'], but found {nan_policy}")
    if nan_policy == 'raise':
        if np.any(np.isnan(x)) or np.any(np.isnan(y)):
            raise ValueError('Found nan in input')
    if ax is None:
        ax = plt.gca()
    if 'alpha' not in shade_args:
        shade_args['alpha'] = 0.5

    allowed_errors = ['stderr','std']
    if err_method not in allowed_errors:
        raise ValueError(f'err_method must be one of {allowed_errors}, but found {err_method}')
    if nan_policy == 'omit':
        y_mean = np.nanmean(y, axis=1)
        if err_method == 'stderr':
            y_err = np.nanstd(y, axis=1) / np.sqrt(y.shape[1])
        elif err_method == 'std':
            y_err = np.nanstd(y, axis=1)
    else:
        y_mean = y.mean(1)
        if err_method == 'stderr':
            y_err = y.std(1) / np.sqrt(y.shape[1])
        elif err_method == 'std':
            y_err = y.std(1)
        
    ax.plot(x, y_mean, **plt_args)
    ax.fill_between(x, y_mean-y_err, y_mean+y_err, **shade_args)
    

def hierarchicalclusterplot(data, axes=None, varnames=None, cmap='bwr', n_clusters=2):
    '''
    Parameters
    ----------
    data : shape (n_samples, n_features)
    axes : array of length 2 containing matplotlib axes (optional)
        axes[0] will be for the dendrogram and axes[1] will be for the data
    varnames : list of strings, length must = n_features, default=None
        variable names which will be printed as yticklabels on the data plot
    cmap : string, default='bwr'
        colormap for the data plot
    n_clusters : int, default=2
        number of clusters which will be used when computing cluster labels that are returned
    
    Returns
    -------
    cluster_dict : dict output from scipy.cluster.hierarchy.dendrogram
    cluster_labels : np.array
        cluster labels from sklearn.cluster.AgglomerativeClustering
    '''
    if axes is None:
        _, axes = plt.subplots(2,1,figsize=(10, 7), gridspec_kw={'height_ratios': [2.5,1]})
        
    dend = shc.dendrogram(shc.linkage(data, method='ward'), show_leaf_counts=False, ax=axes[0], get_leaves=True, no_labels=True)

    axes[0].set_yticks([])

    leaves = dend['leaves']

    cluster = AgglomerativeClustering(n_clusters=n_clusters, affinity='euclidean', linkage='ward')
    cluster_labels = cluster.fit_predict(data)

    if cmap=='bwr':
        mm1 = np.abs(data.reshape((-1)).min())
        mm2 = np.abs(data.reshape((-1)).max())
        mm = max([mm1, mm2])
        axes[1].imshow(data[leaves,:].T, cmap=cmap, aspect=4, vmin=-mm, vmax=mm, interpolation='none')
    else:
        axes[1].imshow(data[leaves,:].T, cmap=cmap, aspect=4, interpolation='none')
    if varnames:
        axes[1].set_yticks([i for i in range(len(varnames))])
        axes[1].set_yticklabels(varnames, fontsize=8)

    axes[1].set_xticks([])

    plt.tight_layout()
    plt.show()
    
    return dend, cluster_labels


def imSTRF(coef, tmin=None, tmax=None, freqs=None, ax=None, smooth=True):
    '''
    Plot STRF weights as image. Weights are automatically centered at 0
    so the center of the colormap is 0.
    
    Parameters
    ----------
    coef : np.array, shape (freq, lag)
        STRF weights.
    tmin : float, optional
        Time of first lag (first column)
    tmax : float, optional
        Time of final lag (last column)
    freqs : list or array-like, length=2, optional
        Frequency of lowest and highest frequency bin in STRF.
    ax : plt.Axes, optional
        Axes to plot on.
    smooth : bool, default=True
        Whether or not to smooth the STRF image. Smoothing is
        done with 'gouraud' shading in plt.pcolormesh().
    '''
    
    if ax is None:
        ax = plt.gca()
    
    if tmin is not None and tmax is not None:
        delays_sec = np.linspace(tmin, tmax, coef.shape[1])
        lag_string = 'Lag (s)'
    else:
        delays_sec = np.arange(0, coef.shape[1])
        lag_string = 'Lag (samples)'
        
    ax.set_xlabel(lag_string)
    
    freqs_ = np.arange(0, coef.shape[0])
    ax.set_ylabel('Frequency')
        
    if smooth:
        kwargs = dict(vmax=np.abs(coef).max(), vmin=-np.abs(coef).max(),
                  cmap='bwr', shading='gouraud')
    else:
        kwargs = dict(vmax=np.abs(coef).max(), vmin=-np.abs(coef).max(),
                  cmap='bwr')
        
    ax.pcolormesh(delays_sec, freqs_, coef, **kwargs)
    
    if freqs is not None:
        yticks = ax.get_yticks()
        ax.set_yticks([0, coef.shape[0]-1])
        ax.set_yticklabels([freqs[0], freqs[-1]])



import numpy as np
from scipy.ndimage import gaussian_filter1d
try:
    import torch
except:
    pass # soft requirement, since may not be needed 

from ..stats import fratio
from ..utils import _parse_outstruct_args
from ..out_struct import OutStruct

def get_label_change_points(x):
    '''
    Find the indices where x changes from one categorical (integer) value to another.
    For example, if x is [0, 0, 0, 1, 1, 3, 3], locations returned will be [3, 5]
    
    Parameters
    ----------
    x : np.ndarray of shape (time,)
        Label vector over time.
    
    Returns
    -------
    locs : array of shape (n_changes, )
        Indices where labels change.
    labels : array of shape (n_changes, )
        New label (from input `x`) after each transition.
    prior_labels : array of shape (n_changes, )
        Old label (from input `x`) before each transition.
    '''
    
    diff_x = np.concatenate([np.array([0]), np.diff(x)])
    locs = np.argwhere(diff_x!=0).squeeze()
    labels = x[locs]
    labels_prior = x[locs-1]
    return locs, labels, labels_prior

def segment_around_label_transitions(outstruct=None, data=None, labels=None, prechange_samples=50, postchange_samples=300, elec_lag=None):
    '''
    Cut x around the transition points given by the changes in the labels.
    
    Parameters
    ----------
    outstruct : OutStruct, optional
        OutStruct containing data to be normalized in one of the field. If not given, must give
        the data to be normalized directly as the ``data`` argument. 
    data : list or array-like, or a string specifying a field of the outstruct
        Data to be segmented based on the data[i] is shape (time, electrode, *, ...)
    labels : string, or list of np.ndarrays or array-like with the same length as data, or a tuple of lists or array-likes each the same length as data
        If a string, specifies a field of the oustruct containing labels to use for
        segmenting. If a single list or np.ndarray (not a tuple), then labels[i]
        is shape (time, ) giving integer label of each trial.
        If a tuple, then labels[0][i] is shape (time, ) giving integer label of each sample, and the
        transition points of these labels are used to segment. The other lists of labels (like labels[1])
        are not used for segmentation but their values surrounding the transition point are returned in a tuple of labels
    prechange_samples : int, default=50
        Number of samples to include in each segment that come before a label transition.
    postchange_samples : int, default=300
        Number of samples to include in each segment that come after a label transition.
    elec_lag : list or array-like, default=None
        Provides the lag (in samples) of each electrode, must be length=n_electrodes=data[i].shape[1].
        Only used if not None. This can be computed with many methods, for example, using
        ``nl.segmentation.electrode_lags_fratio``.
        
    Returns
    -------
    segments : array of shape (n_segments, time, electrode, *, ...)
        Segments cut from data surrounding label transition points.
    labels : array-like
        If input labels is just a list, then is of shape (n_segments,) providing new label after transition point
        If input labels is a tuple of lists, then this is a tuple of array-likes, where the first is the
        same as described above, and the others are arrays of shape (n_segments, time)
            
    prior_labels : array-like, shape (n_segments,)
        Gives label that came before the transition in the main labels array used for segmentation

    Examples
    --------
    >>> from naplib.segmentation import segment_around_label_transitions
    >>> # In this example, we imagine we only have 1 trial of data, so all the lists of data and labels
    >>> # are length 1. In practice, they could be any length though.
    >>> arr = np.arange(20).reshape(10,2) # array of shape (time, features/electrodes)
    >>> arr
    array([[ 0,  1],
           [ 2,  3],
           [ 4,  5],
           [ 6,  7],
           [ 8,  9],
           [10, 11],
           [12, 13],
           [14, 15],
           [16, 17],
           [18, 19]])
    >>> # use a label of categorical values to segment the array based on
    >>> # transitions in the categorical label
    >>> label = np.array([0,0,1,1,1,0,0,3,3,3])
    >>> segments, labels, prior_labels = segment_around_label_transitions(data=[arr], labels=[label],
    ...                                                                   prechange_samples=0,
    ...                                                                   postchange_samples=3)
    >>> segments
    array([[[ 4,  5],
            [ 6,  7],
            [ 8,  9]],
           [[10, 11],
            [12, 13],
            [14, 15]],
           [[14, 15],
            [16, 17],
            [18, 19]]])
    >>> labels, prior_labels
    (array([1, 0, 3]), array([0, 1, 0]))
    >>> # We can also get the full segment value of these labels or another set of
    >>> # labels around the transition, rather than just the single values of the
    >>> # labels at the transition point, simply by passing other label vectors
    >>> # that we want to be segmented as the 2nd through nth value in a tuple of labels
    >>> label2 = np.array([0,1,2,3,4,5,6,7,8,9]) # another set of labels of interest
    >>> label_tuple = ([label], [label], [label2])
    >>> segments, labels, prior_labels = segment_around_label_transitions(data=[arr], labels=label_tuple, prechange_samples=0, postchange_samples=3)
    >>> labels # this time we get a tuple of 3 labels. The first is the same as before, and the others are fully segmented 
    (array([1, 0, 3]),
     array([[1, 1, 1],
            [0, 0, 3],
            [3, 3, 3]]),
     array([[2, 3, 4],
            [5, 6, 7],
            [7, 8, 9]]))


    '''
    
    x, labels = _parse_outstruct_args(outstruct, data, labels)
    
    if isinstance(labels, tuple):
        labels, other_labels = labels[0], labels[1:]
        return_multiple_labels = True
    else:
        other_labels = (labels,)
        return_multiple_labels = False
    
    segments = []
    new_labels = []
    prior_labels = []
    other_labels_to_return = [[] for _ in range(len(other_labels))]
    for i, tmp_unpack in enumerate(zip(x, labels, *other_labels)):
        x_i, labels_i, other_labels_i = tmp_unpack[0], tmp_unpack[1], tmp_unpack[2:]
        if not isinstance(x_i, np.ndarray) and isinstance(x_i, torch.Tensor): # only check if a torch Tensor if not np array
            x_i = x_i.numpy()
        label_changepoints, labels_at_changepoints, labels_before_changepoints = get_label_change_points(labels_i)
        
        label_changepoints = label_changepoints.astype('int')
        labels_before_changepoints = labels_before_changepoints.astype('int')
        
        for i_c, (change_point, new_lab, prior_lab) in enumerate(zip(label_changepoints, labels_at_changepoints, labels_before_changepoints)):
            if change_point > prechange_samples and change_point+postchange_samples <= x_i.shape[0]:
                if elec_lag is not None:
                    tmp_x_i_region = np.array([x_i[change_point-prechange_samples+elec_lag[elec_idx]:change_point+postchange_samples+elec_lag[elec_idx], elec_idx] for elec_idx in range(len(elec_lag))])
                    segments.append(tmp_x_i_region.transpose())
                else:
                    segments.append(x_i[change_point-prechange_samples:change_point+postchange_samples])
                new_labels.append(new_lab)
                prior_labels.append(prior_lab)

                for i, other_labs in enumerate(other_labels_i):
                    other_labels_to_return[i].append(other_labs[change_point-prechange_samples:change_point+postchange_samples])
                    
    if return_multiple_labels:
         return np.array(segments), (np.array(new_labels), *[np.array(tt) for tt in other_labels_to_return]), np.array(prior_labels)
    else:
        return np.array(segments), np.array(new_labels), np.array(prior_labels)
    
def electrode_lags_fratio(outstruct=None, data=None, labels=None, max_lag=20, return_fratios=False):
    '''
    Compute lags of each electrode based on peak of f-ratio to a given label, such as phoneme labels.
    The data is segmented around onset transitions in the labels, and an electrode's lag is defined
    as the peak of the f-ratio after the transition.
    
    Parameters
    ----------
    outstruct : OutStruct, optional
        OutStruct containing data to be normalized in one of the field. If not given, must give
        the data to be normalized directly as the ``data`` parameter. 
    data : string, or list or array-like
        Electrode data to use to compute lags with the fratio.
        If a string, must specify a field of the ``outstruct`` parameter. If a list or array-like,
        data[i] is shape (time, n_electrodes, *, ...)
    labels : string, or list or array-like
        Labels over time for each trial. The onset changes of these labels will
        be used to segment the data and compute the f-ratio after the transition.
        If a string, must specify a field of the outstruct. If a list or array-like, then
        labels[i] is of shape (time,)
    max_lag : int, default=20
        Maximum lag to look for in the f-ratio (in samples).
    return_fratios : bool, default=False
        Whether or not to return smoothed_fratios along with lags in a tuple. Default False.
    
    Returns
    -------
    lags : np.ndarray, shape=(n_electrodes,)
        Lag of each electrode, in samples
    smoothed_fratios : np.ndarray, shape=(n_electrodes, time)
        Fratio of each electrode over time, after gaussian smoothing.
        Only returned if ``return_fratios=True``
    '''
        
    segments, labels, _ = segment_around_label_transitions(outstruct=outstruct, data=data, labels=labels, prechange_samples=0, postchange_samples=max_lag)
    
    fratios_lags = fratio(segments.transpose((2,1,0)), labels, elec_mode='individual')
    
    fratios_lags_smooth = gaussian_filter1d(fratios_lags, 0.5, mode='constant')

    lags = fratios_lags_smooth.argmax(-1)
    
    if return_fratios:
        return lags, fratios_lags_smooth
    else:
        return lags
    



from .aligner import Aligner
from .alignment import get_phoneme_label_vector
from .alignment import get_word_label_vector
from .alignment import create_wrd_dict
from .auditory_spectrogram import auditory_spectrogram

__all__  = ['Aligner','get_phoneme_label_vector','get_word_label_vector','create_wrd_dict','auditory_spectrogram']


from .segmentation import get_label_change_points
from .segmentation import segment_around_label_transitions
from .segmentation import electrode_lags_fratio

__all__  = ['get_label_change_points','segment_around_label_transitions','electrode_lags_fratio']



from copy import deepcopy
import numpy as np

from ..out_struct import OutStruct
from ..utils import _parse_outstruct_args

def normalize(outstruct=None, data='resp', axis=0, method='zscore'):
    '''Normalize data over multiple trials. If you are trying to normalize a single matrix,
    it must be passed in inside a list, or with an extra first dimension.
    
    Parameters
    ----------
    outstruct : naplib.OutStruct object, optional
        OutStruct containing data to be normalized in one of the field. If not given, must give
        the data to be normalized directly as the ``data`` argument. 

    data : string if outstruct given, or a list of np.ndarrays or a multidimensional np.ndarray
        Data to normalize. If a string, it must specify one of the fields of the outstruct
        provided in the first argument. If a multidimensional array, first dimension
        indicates the trial/instances which will be concatenated over to compute
        normalization statistics.
    
    axis : int, default=-1
        Axis of the array to normalize over.
    
    method : string, default='zscore'
        Method of normalization. Must be one of ['zscore','center'].
        'center' only centers the data, while 'zscore' also scales by standard deviation
    
    Returns
    -------
    normalized_data : list of np.ndarrays
    
    '''
    
    data = _parse_outstruct_args(outstruct, data)
    
    if method not in ['zscore', 'center']:
        raise ValueError(f"Bad method input. method must be one of ['zscore', 'center'], but found {method}")

    if isinstance(data, np.ndarray):
        data = [d for d in data]
        axis -= 1 # since we got rid of the first axis by putting into a list, need to change this
    elif not isinstance(data, list) or not isinstance(data[0], np.ndarray):
        raise TypeError(f'data found is not either np.ndarray, or list or arrays, but found {type(data)}')

    concat_data = np.concatenate(data, axis=axis)
    
    center_val = np.mean(concat_data, axis=axis, keepdims=True)
    if method=='zscore':
        std_val = np.std(concat_data, axis=axis, keepdims=True)
    
    for i, tmp in enumerate(data):
        if method=='zscore':
            data[i] = (tmp - center_val) / std_val
        elif method=='center':
            data[i] = tmp - center_val

    return data




# Copyright (c) 2011-2014 Kyle Gorman and Michael Wagner
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Serialization/deserialization functionality using `shutil`
"""

import os

from tempfile import mkdtemp
from shutil import copy, rmtree, make_archive, unpack_archive

from utilities import mkdir_p


# default format for output
FORMAT = "zip"


class Archive(object):

    """
    Class representing data in a directory or archive file (zip, tar, 
    tar.gz/tgz)
    """

    def __init__(self, source, is_tmpdir=False):
        if os.path.isdir(source):
            self.dirname = os.path.abspath(source)
            self.is_tmpdir = is_tmpdir  # trust caller
        else:
            base = mkdtemp(dir=os.environ.get("TMPDIR", None))
            unpack_archive(source, base)
            (head, tail, _) = next(os.walk(base))
            if not tail:
                raise ValueError("'{}' is empty.".format(source))
            if len(tail) > 1:
                raise ValueError("'{}' is a bomb.".format(source))
            self.dirname = os.path.join(head, tail[0])
            self.is_tmpdir = True  # ignore caller

    @classmethod
    def empty(cls, head):
        """
        Initialize an archive using an empty directory
        """
        base = mkdtemp(dir=os.environ.get("TMPDIR", None))
        source = os.path.join(base, head)
        mkdir_p(source)
        return cls(source, True)

    def add(self, source):
        """
        Add file into archive
        """
        copy(source, self.dirname)

    def __repr__(self):
        return "{}(dirname={!r})".format(self.__class__.__name__,
                                         self.dirname)

    def dump(self, sink, archive_fmt=FORMAT):
        """
        Write archive to disk, and return the name of final archive
        """
        return make_archive(sink, archive_fmt,
                            *os.path.split(self.dirname))

    def __del__(self):
        if self.is_tmpdir:
            rmtree(self.dirname)



import os
from os.path import isfile, join, isdir, dirname
import sys
import unicodedata
import string

import numpy as np
from scipy.io.wavfile import write as write_wavfile

from .alignment import create_wrd_dict, get_phoneme_label_vector, get_word_label_vector
from ..utils import _parse_outstruct_args
from ..out_struct import OutStruct


class Aligner():
    '''
    This class performs phoneme and word alignment using audio files
    and matching text files containing scripts. If words in the texts do not
    appear in the dict file, you will need to add them to a dict file and specify
    it as ``dictionary_file``.

    Please see the :ref:`alignment example notebooks <alignment examples>` for more detailed
    tutorials which show how to align text and audio data and analyze the output.

    Note
    ----
    Several extra packages are required to perform alignment. Please follow the
    installation instructions for ``HTK`` and ``sox`` for your system before
    using alignment. Additionally, you will need to install
    `pyyaml <https://pypi.org/project/PyYAML/>`_ as well as
    `TextGrid <https://pypi.org/project/TextGrid/>`_ for the Aligner to work. These are
    not required dependencies of naplib-python, so they must be installed separately.
    
    Parameters
    ----------
    output_dir : string, path-like
            Directory to put output files in, such as .phn, .wrd., and .TextGrid files.
    dictionary_file : string, path-like, optional
        Path to a dictionary file (e.g. eng.dict) which contains phonemes for
        all words in corpus. If not provided, will use the default eng.dict.
        For an example file, see
        `ProsodyLab's eng.dict <https://github.com/prosodylab/Prosodylab-Aligner/blob/master/eng.dict>`_ 
    tmp_dir : string, path-like, optional
        Directory to hold temporary files that are created. If not provided, creates
        a folder called `data_/` in the current working directory and uses that.
    verbose : int, default=1
            Integer specifying the level of verbosity of print statements and messages
            while the function is running. 0 indicates no output, and higher values
            generate more output.
    '''
    def __init__(self, output_dir, dictionary_file=None, tmp_dir=None, verbose=1):

        os.makedirs(output_dir, exist_ok=True)
        if tmp_dir is None:
            tmp_dir = 'data_/'
            # check if this folder already exists, in which case throw an error
            # so we don't overwrite a folder by default.
            if isdir(tmp_dir):
                raise ValueError(f'No tmp_dir was provided, but could not use '
                    'the default "data_/" because a folder with that name '
                    'already exists in the current path. Please remove that '
                    'directory or explicitly specify the tmp_dir parameter.')
        self.output_dir = output_dir
        self.tmp_dir = tmp_dir
        self.filedir_ = dirname(__file__)
        if dictionary_file is None:
            dictionary_file = join(self.filedir_, 'eng.dict')
        self.dictionary_file = dictionary_file
        self.verbose = verbose

        try:
            import yaml
        except Exception as e:
            raise Exception('Missing package pyyaml which is required for alignment. Please '
                'install it with "pip install pyyaml"')
        try:
            import textgrid
        except Exception as e:
            raise Exception('Missing package TextGrid which is required for alignment. Please '
                'install it with "pip install TextGrid"')


    def _remove_nonword_characters_and_punctuation_and_capitalize(self, s):
        exclude = set(string.punctuation)
        exclude.remove("'")
        s = ''.join(ch for ch in s if ch not in exclude)
        # s = s.translate(str.maketrans('', '', string.punctuation))
        s = s.upper()
        return s

    def _convert_text_to_ascii(self, name, root):
        new_name = name.replace('.txt', '.lab')
        new_folder = self.tmp_dir

        unicode_file = open(os.path.join(root, name))
        unicode_data = unicode_file.read() #.decode(input_codec)
        unicode_data = self._remove_nonword_characters_and_punctuation_and_capitalize(unicode_data)
        ascii_data = unicodedata.normalize('NFKD', unicode_data).encode('ascii','ignore')
        ascii_file = open(os.path.join(new_folder, new_name), 'wb')
        ascii_file.write(ascii_data)

    def align(self, outstruct=None, name='name', sound='sound',
              soundf='soundf', transcript='transcript',
              dataf='dataf', length='length'):
        '''
        Perform alignment across a set of paired audio-text files stored
        in fields of an OutStruct. This function will create a set of
        .TextGrid files, as well as corresponding .phn and .wrd
        files in the output_dir which describe the timing of phonemes and
        words within each audio. These files can be used in conjunction
        with the other functions in `naplib.alignment`, such as
        ``get_phoneme_label_vector`` and ``get_word_label_vector``,
        which take these files as input. This function will automatically
        use ``naplib.alignment.get_phoneme_label_vector`` and
        ``naplib.alignment.get_word_label_vector`` to produce phoneme and
        word label vectors for each stimulus which can be placed into the
        OutStruct and further analyzed.

        This function is essentially equivalent to storing audio and text
        in directories and using ``Aligner.align_files`` followed by
        ``Aligner.get_label_vecs_from_files``.

        Parameters
        ----------
        outstruct : OutStruct
            OutStruct containing the data to align. It must contain the
            following fields. 
        name : string or list of strings, default='name'
            If a string, specifies a field of the outstruct which contains
            the name for each trial. Otherwise, a list of strings specifies
            the name for each trial.
        sound : string or list of np.ndarrays, default='sound'
            If a string, specifies a field of the outstruct which contains
            the sound waveform for each trial. Otherwise, a list of np.ndarrays
            specifies the waveform for each trial.
        soundf : string, integer, or list of integers, default='soundf'
            If a string, specifies a field of the outstruct which contains
            the sampling rate for each trial. Otherwise, a list of integers
            specifies the sampling rate for each trial, or a single integer gives the
            sampling rate for all trials.
        transcript : string or list of strings, default='transcript'
            If a string, specifies a field of the outstruct which contains
            the transcript text for each trial. Otherwise, a list of strings
            specifies the transcript text for each trial.
        dataf : string, integer, or list of integers, default='dataf'
            If a string, specifies a field of the outstruct which contains
            the desired sampling rate of the output. Otherwise, a list of integers
            specifies the Desired sampling rate of the output for each trial, or
            a single integer gives the desired sampling rate of the output
            for all trials.
        length : string or list of integers, default='length'
            If a string, specifies a field of the outstruct which contains
            the desired output length (in samples) for each trial. Otherwise,
            a list of integers specifies the desired output length (in samples)
            for each trial.

        Returns
        -------
        alignment_outstruct: OutStruct
            OutStruct containing all alignment information, with all the fields
            described by the return values below. 
        phn_labels : list of np.ndarrays
            Phoneme label vector for each trial. alignment_outstruct['phn_labels'][i]
            is a np.ndarray of shape (time,) and sampling rate dataf[i].
        manner_labels : list of np.ndarrays
            Manner of articulation label vector for each trial.
            alignment_outstruct['manner_labels'][i]
            is a np.ndarray of shape (time,) and sampling rate dataf[i].
        wrd_labels : list of np.ndarrays
            Word label vector for each trial. alignment_outstruct['wrd_labels'][i]
            is a np.ndarray of shape (time,) and sampling rate dataf[i].
        phn_label_list : list of lists of strings
            Phoneme label list returned by ``naplib.alignment.get_phoneme_label_vector``,
            so alignment_outstruct['phn_label_list'][i] is a list of phonemes, where the
            index of a given phoneme in the list encodes that phoneme's label in ``phn_labels``.
        manner_label_list : list of lists of strings
            Manner of articulation label list returned by ``naplib.alignment.get_phoneme_label_vector``,
            so alignment_outstruct['manner_label_list'][i] is a list of manners, where the
            index of a given manner in the list encodes that manner's label in ``manner_labels``.
        wrd_dict : dict
            Dictionary of word:int (key:value) pairs for all the words in the corpus
            of files in the directory, created by ``naplib.alignment.create_wrd_dict``
            So, alignment_outstruct['wrd_dict'][i] is a dictionary
            which maps a word to its integer value as it is represented in ``wrd_labels``.

        Note
        ----
        This function will produce the following files in the output_dir to aid in
        its running.

        | working directory
        | └── output_dir
        | │   └── trial1.phn
        | │   └── trial1.wrd
        | │   └── trial1.TextGrid
        | │   └── trial2.phn
        | │   └── trial2.wrd
        | │   └── trial2.TextGrid
        '''

        names, sounds, soundf, transcripts, dataf, lengths = _parse_outstruct_args(outstruct,
                                                                                   name,
                                                                                   sound,
                                                                                   soundf,
                                                                                   transcript,
                                                                                   dataf,
                                                                                   length)

        # Write sounds to wav files in tmp folder and text to .txt files
        audio_dir = join(self.tmp_dir, 'tmp_sounds')
        os.makedirs(audio_dir, exist_ok=True)
        text_dir = join(self.tmp_dir, 'tmp_text')
        os.makedirs(text_dir, exist_ok=True)
        for name, soundwave, soundf_, script in zip(names, sounds, soundf, transcripts):
            fname_wav = join(audio_dir, f'{name}.wav')
            write_wavfile(fname_wav, int(soundf_), soundwave)
            fname_txt = join(text_dir, f'{name}.txt')
            with open(fname_txt, "w") as text_file:
                n = text_file.write(script)
                text_file.close()

        # Align text and audio from files
        self.align_files(audio_dir, text_dir)

        # Get the label vectors from the alignment files
        return self.get_label_vecs_from_files(outstruct=outstruct, name=names,
                                  dataf=dataf, length=lengths,
                                  befaft=np.array([0, 0]))


    def align_files(self, audio_dir, text_dir):
        '''
        Perform alignment across a set of paired audio-text files stored
        in directories. This function will create a set of .TextGrid files,
        as well as corresponding .phn and .wrd
        files in the output_dir which describe the timing of phonemes and
        words within each audio. These files can be used in conjunction
        with the other functions in `naplib.alignment`, such as
        ``get_phoneme_label_vector`` and ``get_word_label_vector``,
        which take these files as input.

        Parameters
        ----------
        audio_dir : string, path-like
            Directory containing audio files (.wav).
        text_dir : string, path-like
            Directory containing text files (.txt) with matching names
            to the files in ``audio_dir``.

        Note
        ----
        The directory structure containing audios and matching text
        files must be correct in order to properly perform alignment.
        See below for what the directory layout should look like
        before running this function.

        | working directory
        | ├── audio_dir
        | │   ├── file1.wav
        | │   ├── file2.wav
        | └── text_dir
        | │   └── file1.txt
        | │   └── file2.txt 

        After running this function, the directory layout will look
        like this:

        | working directory
        | ├── audio_dir
        | │   ├── file1.wav
        | │   ├── file2.wav
        | └── text_dir
        | │   └── file1.txt
        | │   └── file2.txt
        | └── output_dir
        | │   └── file1.phn
        | │   └── file1.wrd
        | │   └── file1.TextGrid
        | │   └── file2.phn
        | │   └── file2.wrd
        | │   └── file2.TextGrid
        '''
        import textgrid

        if self.verbose >= 1:
            print(f'Resampling audio and putting in {self.tmp_dir} directory...')

        resample_path = join(self.filedir_, 'resample.sh')

        # resample the audios to 16000 and put them in the tmp data folder
        os.system(f'{resample_path} -s 16000 -r {audio_dir} -w {self.tmp_dir}')

        if self.verbose >= 1:
            print(f'Converting text files to ascii in {self.tmp_dir} directory...')

        for root, dirs, files in os.walk(text_dir, topdown=False):
            for name in files:
                if '.txt' in name:
                    self._convert_text_to_ascii(name, root)

        if self.verbose >= 1:
            print('Performing alignment...')

        # perform alignment using ProsodyLab-Aligner
        sys.path.insert(1, self.filedir_)
        prosodylab_main_file = join(self.filedir_, 'prosodylab_aligner/__main__.py')
        eng_zip_file = join(self.filedir_, 'eng.zip')
        os.system(f'python3 {prosodylab_main_file} -a {self.tmp_dir} -d {self.dictionary_file} -r {eng_zip_file}')
        sys.path.remove(self.filedir_)

        if self.verbose >= 1:
            print(f'Converting .TextGrid files to .phn and .wrd in {self.output_dir}')

        # Convert textgrid files to .phn and .wrd files in output_dir
        for root, dirs, files in os.walk(self.tmp_dir, topdown=False):
            for name in files:
                if '.TextGrid' in name:

                    # print(f'looking at {os.path.join(root, name)}')

                    # copy TextGrid file to output_dir so they are saved
                    os.system(f'cp {join(root, name)} {join(self.output_dir, name)}')

                    new_phn_name = name.replace('.TextGrid', '.phn')
                    new_wrd_name = name.replace('.TextGrid', '.wrd')

                    tg = textgrid.TextGrid.fromFile(join(root, name))
                    phones = tg[0]
                    words = tg[1]

                    # write phn file

                    phn_file = open(os.path.join(self.output_dir, new_phn_name), 'w')

                    for phone_seg in phones:
                        if phone_seg.mark == "":
                            phone_seg.mark = "sp"
                        if phone_seg.mark != "sil":
                            print(f"{phone_seg.minTime} {phone_seg.maxTime} {phone_seg.mark}", file=phn_file)

                    phn_file.close()

                    # write wrd file

                    wrd_file = open(os.path.join(self.output_dir, new_wrd_name), 'w')

                    for word_seg in words:
                        if word_seg.mark != "sil":
                            print(f"{word_seg.minTime} {word_seg.maxTime} {word_seg.mark}", file=wrd_file)

                    wrd_file.close()


        if self.verbose >= 1:
            print('Finished creating alignment files.')

    def get_label_vecs_from_files(self, outstruct=None, name='name',
                                  dataf='dataf', length='length',
                                  befaft='befaft'):
        '''

        Parameters
        ----------
        outstruct : OutStruct
            OutStruct containing the data to align. It must contain the
            following fields. 
        name : string or list of strings, default='name'
            If a string, specifies a field of the outstruct which contains
            the name for each trial. Otherwise, a list of strings specifies
            the name for each trial.
        dataf : string, integer, or list of integers, default='dataf'
            If a string, specifies a field of the outstruct which contains
            the desired sampling rate of the output. Otherwise, a list of integers
            specifies the Desired sampling rate of the output for each trial, or
            a single integer gives the desired sampling rate of the output
            for all trials.
        length : string or list of integers, default='length'
            If a string, specifies a field of the outstruct which contains
            the desired output length (in samples) for each trial. Otherwise,
            a list of integers specifies the desired output length (in samples)
            for each trial.
        befaft : string or list of np.ndarrays, or a single np.ndarray, default='befaft'
            If a string, specifies a field of the outstruct which contains
            the before and after time (in sec) for each trial. Otherwise,
            a list should contain the befaft period for each trial, and a single
            np.ndarray of length 2 specifies the befaft period for all trials. For
            example, befaft=np.array([0.5, 0.5]) indicates that for each trial, the
            wav file which was used to produce the alignment is 0.5 seconds shorter
            at the beginning and 0.5 seconds shorter at the end than the desired
            output length.

        Returns
        -------
        alignment_outstruct: OutStruct
            OutStruct containing all alignment information, with all the fields
            described by the return values below. 
        phn_labels : list of np.ndarrays
            Phoneme label vector for each trial. alignment_outstruct['phn_labels'][i]
            is a np.ndarray of shape (time,) and sampling rate dataf[i].
        manner_labels : list of np.ndarrays
            Manner of articulation label vector for each trial.
            alignment_outstruct['manner_labels'][i]
            is a np.ndarray of shape (time,) and sampling rate dataf[i].
        wrd_labels : list of np.ndarrays
            Word label vector for each trial. alignment_outstruct['wrd_labels'][i]
            is a np.ndarray of shape (time,) and sampling rate dataf[i].
        phn_label_list : list of lists of strings
            Phoneme label list returned by ``naplib.alignment.get_phoneme_label_vector``,
            so alignment_outstruct['phn_label_list'][i] is a list of phonemes, where the
            index of a given phoneme in the list encodes that phoneme's label in ``phn_labels``.
        manner_label_list : list of lists of strings
            Manner of articulation label list returned by ``naplib.alignment.get_phoneme_label_vector``,
            so alignment_outstruct['manner_label_list'][i] is a list of manners, where the
            index of a given manner in the list encodes that manner's label in ``manner_labels``.
        wrd_dict : dict
            Dictionary of word:int (key:value) pairs for all the words in the corpus
            of files in the directory, created by ``naplib.alignment.create_wrd_dict``
            So, alignment_outstruct['wrd_dict'][i] is a dictionary
            which maps a word to its integer value as it is represented in ``wrd_labels``.

        Note
        ----
        This function requires that the following files ALREADY exist in the aligner's
        output_dir.

        | working directory
        | └── output_dir
        | │   └── trial1.phn
        | │   └── trial1.wrd
        | │   └── trial1.TextGrid
        | │   └── trial2.phn
        | │   └── trial2.wrd
        | │   └── trial2.TextGrid
        '''

        names, dataf, lengths, befafts = _parse_outstruct_args(outstruct, name, dataf, length, befaft)

        wrd_dict = create_wrd_dict(self.output_dir)

        alignment_results = []

        if self.verbose >= 1:
            print(f'Creating label vectors for phonemes, manner of articulation, and words.')

        for n in range(len(outstruct)):

            this_trial_result = {}
    
            # filenames for the .phn and .wrd files
            filename_phn = join(self.output_dir, f'{names[n]}.phn')
            filename_wrd = join(self.output_dir, f'{names[n]}.wrd')
            
            # desired length of the output label vector
            length = lengths[n]
            
            # sampling rate of our data
            fs = dataf[n]
            
            # before-after period for our data is 0 since we are using an outstruct where the
            # durations of sound and output should already be matched
            befaft = befafts[n]
            
            # compute label vectors for phonemes, manner of articulation, and words, for this trial
            label_vec_phn, phn_label_list = get_phoneme_label_vector(filename_phn, length, fs, befaft, return_label_lists=True)
            label_vec_manner, manner_label_list = get_phoneme_label_vector(filename_phn, length, fs, befaft, mode='manner', return_label_lists=True)
            label_vec_wrd = get_word_label_vector(filename_wrd, length, fs, befaft, wrd_dict=wrd_dict)

            this_trial_result['phn_labels'] = label_vec_phn
            this_trial_result['manner_labels'] = label_vec_manner
            this_trial_result['wrd_labels'] = label_vec_wrd
            this_trial_result['phn_label_list'] = phn_label_list
            this_trial_result['manner_label_list'] = manner_label_list
            this_trial_result['wrd_dict'] = wrd_dict

            alignment_results.append(this_trial_result)
            
        # Add the computed label vectors to the outstruct
        return OutStruct(alignment_results, strict=False)





import warnings
from statsmodels.tools.sm_exceptions import ConvergenceWarning
import statsmodels.formula.api as smf
from statsmodels.regression.mixed_linear_model import MixedLM
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


class LinearMixedEffectsModel():
    '''
    A linear mixed effects model which can be used for main effect plots.
    A mixed effects model is useful when the data has some non-independence
    whose effect should be accounted for separately from the effects of
    the independent variables. For example, if half of the data comes
    from subject A and the other half from subject B, and you want to
    know the main effects of other predictor features.
    
    Parameters
    ----------
    alpha : float, default=0.05
        Alpha level for confidence interval.
    zscore_x : bool, default False
        Whether or not to zscore the columns of X when calling .fit().
        
    Attributes
    ----------
    varnames : list
        List of variable names
    mixedlm : statsmodels.regression.mixed_linear_model.MixedLM object
        Object which performs mixed linear model fitting
    mixedlm_results : statsmodels.regression.mixed_linear_model.MixedLMResults object
        Object which contains outputs from fitting.
    params : pd.DataFrame
        Coefficients for each variable.
    pvalues : pd.DataFrame
        Pvalues for each param.
    conf_int : pd.DataFrame
        Confidence interval around each param.
    rsquared : np.float
        R-squared value.
    
    '''
    def __init__(self, alpha=0.05, zscore_x=False):
        self._alpha = alpha
        self._zscore_x = zscore_x
        self.varnames = []
        self._isfit = False
        
    def fit(self, X, y, random_effect=None, varnames=None):
        '''
        Parameters
        ----------
        X : np.array, shape (num_samples, num_features)
        y : np.array, shape (num_samples,)
        random_effect : np.array, shape (num_samples,), optional
            If given, used as a random effect in the model. For example,
            to use group identity as a random effect, this should be an array
            of integers specifying what group each sample belongs to. The
            values do not matter, only the categorical groups they form.
        varnames : list, optional, default None
            List of variable names, must be length (num_features+1) or
            (num_features + num_features_r + 1), giving the name
            for each feature in X, each feature in random_effect (if given)
            as well as a name for the predicted output in y.
            
        Returns
        -------
            model : returns an instance of self
        '''
        if random_effect is not None:
            if random_effect.ndim == 1:
                random_effect = random_effect[:, np.newaxis]
            elif random_effect.shape[1] > 1:
                raise ValueError(f'Only 1 random effect variable is currently supported, but got {random_effect.shape[1]}')
        
        
        if varnames is None:
            varnames = [f'feature-{i}' for i in range(X.shape[1])]
            varnames.append('y')
        if random_effect is None and len(varnames) != X.shape[1] + 1:
            raise Exception(f'Error: incorrect length of input "varnames". Must include name for each feature in X as well as a name for predicted y.')
        if random_effect is not None and len(varnames) != X.shape[1] + random_effect.shape[1] + 1:
            raise Exception(f'Error: incorrect length of input "varnames". Must include name for each feature in X, each feature in random_effect,'
                            f' as well as a name for predicted y.')
        self.varnames = varnames[:X.shape[1]] + varnames[-1:]
        
        # remove spaces and dashes because they break the smf.mixedlm formula
        varnames_formula = varnames.copy()
        for bad_char in [' ', '-']:
            varnames_formula = [v.replace(bad_char, '') for v in varnames_formula]
        
        warnings.simplefilter('ignore', ConvergenceWarning)
        
        if self._zscore_x:
            X = (X - X.mean(0)) / X.std(0)
            
        self._y = y
        
        if random_effect is None:
            df_tmp = pd.DataFrame(np.concatenate((X, y.reshape(-1,1)), axis=1), columns=varnames_formula)
            formula = f'{varnames_formula[-1]} ~ {varnames_formula[0]}'
            for varname in varnames_formula[1:-1]:
                formula += f' + {varname}'
            mixedlm = smf.mixedlm(formula=formula, data=df_tmp, groups=[1 for _ in range(df_tmp.shape[0])])

        else:
            df_tmp = pd.DataFrame(np.concatenate((X, random_effect, y.reshape(-1,1)), axis=1), columns=varnames_formula)
            formula = f'{varnames_formula[-1]} ~ {varnames_formula[0]}'
            for varname in varnames_formula[1:X.shape[1]]:
                formula += f' + {varname}'
            mixedlm = smf.mixedlm(formula=formula, data=df_tmp, groups=varnames[-2], re_formula='1')

            
        mixedlm_results = mixedlm.fit()
        
        self.mixedlm = mixedlm
        self.mixedlm_results = mixedlm_results
        
        self.params = mixedlm_results.params
        self.pvalues = mixedlm_results.pvalues
        self.conf_int = mixedlm_results.conf_int(alpha=self._alpha)
        
        self._isfit = True
        
        return self
        
    def get_model_params(self):
        '''
        Get model params after fitting.
        
        Returns
        -------
        param_dict : dict   
            dict of model params, pvalues, and confidence intervals for each variable.
        '''
        if not self._isfit:
            raise Exception('Must call .fit() before trying to acces model params.')
        return {'params': np.array(self.params)[1:-1], 'pvalues': np.array(self.pvalues)[1:-1],
                'conf_int': np.array(self.conf_int)[1:-1,:]}
    
    def plot_effects(self, ax=None, plus_minus_colors=None, line_alpha=0.6, markersize=8, center_zero=True, print_ylabels=True):
        '''
        Create main effects plot.
        
        Parameters
        ----------
        ax : pyplot axes, optional, default=gca()
        plus_minus_colors : list or np.array, length=2, default red, blue
            Colors to use for the effect lines when they are significantly positive or negative, respectively.
        line_alpha : float, default=0.06
            alpha value for lines
        markersize : float, default=8
            Size of marker at each effect
        center_zero : bool, default=True
            Whether to put 0 in the center of x-axis.
        print_ylabels : bool, default=True
            Whether to print y labels
        '''
        if ax is None:
            ax = plt.gca()
        
        if plus_minus_colors is None:
            plus_minus_colors = np.array([[255., 44, 0], [97., 164, 205]]) / 255 # blue and red
        
        model_params = self.get_model_params()
        
        
        for effect_idx in range(len(model_params['params'])):
            # check if this effect is significant
            if model_params['params'][effect_idx] > 0 and model_params['conf_int'][effect_idx].min() > 0:
                color_to_use = plus_minus_colors[0]
            elif model_params['params'][effect_idx] < 0 and model_params['conf_int'][effect_idx].max() < 0:
                color_to_use = plus_minus_colors[1]
            else:
                color_to_use = [0.5, 0.5, 0.5]
                
            ax.plot(model_params['conf_int'][effect_idx,:], effect_idx * np.ones_like(model_params['conf_int'][effect_idx-1,:]), color=color_to_use, alpha=line_alpha)
            ax.plot(model_params['params'][effect_idx], np.array([effect_idx]), marker="o", markersize=markersize, markeredgecolor=color_to_use, markerfacecolor=color_to_use)
        
        ax.set_ylim([-0.5, 1.5])
        if center_zero:
            xabs_max = abs(max(ax.get_xlim(), key=abs)) * 1.1
            ax.set_xlim(xmin=-xabs_max, xmax=xabs_max)
        
        ylims = ax.get_ylim()
        ax.set_ylim([-0.5, len(model_params['params'])-0.5])
        ylims = ax.get_ylim()
        ax.vlines(0, ylims[0], ylims[1],'k',linestyles='dashed', linewidth=1, alpha=0.5)

        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        
        if print_ylabels:
            ax.tick_params(top=False, bottom=True, left=False, right=False, labelleft=True, labelbottom=True)
        else:
            ax.tick_params(top=False, bottom=True, left=False, right=False, labelleft=False, labelbottom=True)
        
        ax.set_yticks([nn for nn in range(len(model_params['params']))])
        ax.set_yticklabels(self.varnames[:-1])
        
        return ax
    
    def summary(self):
        '''
        Print summary of model.
        '''
        print(self.mixedlm_results.summary())
        
    @property
    def rsquared(self):
        '''r-squared : float'''
        rss_tot = np.sum(np.square(self._y-self._y.mean()))
        rss_res = np.sum(np.square(self.mixedlm_results.resid))
        return 1.0 - rss_res/rss_tot



from .encoding import dprime
from .encoding import fratio
from .mixedeffectsmodel import LinearMixedEffectsModel
from .pvalues import stars
from .responsive_elecs import responsive_ttest

__all__  = ['dprime','fratio','LinearMixedEffectsModel','stars','responsive_ttest']



from copy import deepcopy
import numpy as np
from scipy.stats import ttest_ind
from mne.stats import fdr_correction

from ..utils import _parse_outstruct_args
from ..out_struct import OutStruct


def responsive_ttest(outstruct=None, resp='resp', befaft='befaft', sfreq='dataf', alpha=0.05, fdr_method='indep', alternative='two-sided', equal_var=True, random_state=None):
    '''
    Identify responsive electrodes by performing a t-test between response
    values during silence (before stimulus) compared to during speech/sound
    (after stimulus onset) [1]_.

    `scipy's ttest_ind <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ttest_ind.html>`_
    is used to perform the t-test. Please see their documentation for more details.

    Parameters
    ----------
    outstruct : OutStruct instance, optional
        OutStruct containing data to be normalized in one of the field.
        If not given, must give the X and y data directly as the ``X``
        and ``y`` arguments. 
    resp : str | list of np.ndarrays or a multidimensional np.ndarray
        Electrode responses to each trial, containing a portion at the
        beginning which is the response to silence before the start
        of the stimulus. Once arranged, each trial should be of
        shape (time, num_channels).
        If a string, it must specify one of the fields of the outstruct
        provided in the first argument. If a multidimensional array,
        first dimension indicates the trial/epochs.
    befaft : str | list of np.ndarrays or a single np.ndarray, default='befaft'
        If a string, specifies a field of the outstruct which contains
        the before and after time (in sec) for each trial. Otherwise,
        a list should contain the befaft period for each trial, and a single
        np.ndarray of length 2 specifies the befaft period for all trials. For
        example, befaft=np.array([0.5, 0.5]) indicates that for each trial,
        the first half second of the `resp` is the responses before the onset
        of the stimulus, and also the final half second is responses for half
        a second after the stimulus ended. If no outstruct is provided, this
        cannot be a string.
    sfreq : str | int, default='dataf'
        The sampling frequency of the responses. If a string, specifies field of
        the outstruct containing the sampling frequency.
    alpha : float, default=0.05
        Error rate.
    fdr_method : str, {'indep', 'negcorr', None}, default='indep'
        Method of correction for multiple comparisons. If 'indep' it implements 
        Benjamini/Hochberg for independent or if 'negcorr' it corresponds
        to Benjamini/Yekutieli. If None, no false discovery rate correction is performed.
    alternative : str, {‘two-sided’, ‘less’, ‘greater’}, default='two-sided'
        The  the alternative hypothesis. Must be one of the following:
        * 'two-sided': the means of the distributions of the response values before
          and during sound are unequal.
        * 'less': the mean of the distribution of response values before stimulus onset
          is less than the mean of the distribution of response values after stimulus onset
        * 'greater': the mean of the distribution of response values before stimulus onset
          is greater than the mean of the distribution of response values after stimulus onset
    equal_var : bool, default=True
        If True, perform a standard independent 2 sample test
        that assumes equal population variances [2]_.
        If False, perform Welch's t-test, which does not assume equal
        population variance [3]_.
    random_state : int, default=None
        Random seed which can be set for reproducibility.

    Returns
    -------
    out : OutStruct | list of np.arrays, same as `resp` input type
        If outstruct was given as input, this is a copy of that
        outstruct with the `resp` field replaced by only the responsive
        electrode data. If `resp` input was given as a list of arrays
        or a 3D array, then this is a list of numpy arrays, each of shape
        (time, new_num_channels)
    stats : dict
        Dictionary containing statistics about responsive elecs, with the
        following keys
        - 'pval' : p-values, shape (num_channels,)
        - 'stat' : test statistic, shape (num_channels,)
        - 'significant': True if null hypothesis was rejected (response is significantly different), False if not, shape (num_channels,)
        - 'alpha': error rate of the test


    References
    ----------
    .. [1] Mesgarani, N., & Chang, E. F. (2012). Selective cortical representation
           of attended speaker in multi-talker speech perception. Nature, 485(7397), 233-236.
    .. [2] https://en.wikipedia.org/wiki/T-test#Independent_two-sample_t-test
    .. [3] https://en.wikipedia.org/wiki/Welch%27s_t-test

    '''

    if fdr_method is not None and fdr_method not in ['indep', 'negcorr']:
        raise ValueError(f"fdr_method should be 'indep' or 'negcorr' but got {fdr_method}")

    if alternative not in ['two-sided', 'less', 'greater']:
        raise ValueError(f"alternative must be one of ['two-sided', 'less', 'greater'] but got {alternative}")

    if isinstance(outstruct, OutStruct):
        return_outstruct = True
        outstruct_copy = deepcopy(outstruct)
        if isinstance(resp, str):
            resp_fieldname = deepcopy(resp) # if this is a string, need to save the name now
        else:
            resp_fieldname = 'resp'
    else:
        return_outstruct = False

    resp, befaft, sfreq = _parse_outstruct_args(outstruct, deepcopy(resp), befaft, sfreq,
                                         allow_different_lengths=True,
                                         allow_strings_without_outstruct=False)

    if isinstance(resp, np.ndarray):
        resp = [r for r in resp]

    if not isinstance(resp, list):
        raise TypeError(f'resp parameter must specify a list of trials or a multidimensional array, but got {type(resp)}')

    if not all([r.ndim == 2 for r in resp]):
        raise ValueError(f'Some response trials are not 2-dimensional (time by channels)')

    # For each trial, pick out the regions before and immediately after stimulus onset
    before_samples = []
    after_samples = []
    pvals = []
    statistics = []
    rng = np.random.default_rng(random_state)
    for t in range(len(resp)):
        bef = int(befaft[t][0] * sfreq[t])
        if bef < 5:
            raise ValueError(f'befaft period is too short, there must be at least 3 samples of response before stimulus onset.')
        before_samples.append(resp[t][:bef])
        after_samples.append(resp[t][bef:3*bef])

    before_samples = np.concatenate(before_samples, axis=0)
    after_samples = np.concatenate(after_samples, axis=0)

    N_retest = 20
    # do the test N_retest times and average the stats
    for _ in range(N_retest):
        before_samples_permuted = rng.permuted(before_samples, axis=0)
        after_samples_permuted = rng.permuted(after_samples, axis=0)
        N_test_samples = min([before_samples_permuted.shape[0]//2, after_samples_permuted.shape[0]//2])
        stat, pval = ttest_ind(before_samples_permuted[:N_test_samples], after_samples_permuted[:N_test_samples], axis=0,
                               equal_var=equal_var, alternative=alternative)
        pvals.append(pval)
        statistics.append(stat)

    statistics = np.array(statistics).mean(0)
    pvals = np.array(pvals).mean(0)

    if fdr_method is not None:
        reject, pval_corrected = fdr_correction(pvals, alpha=alpha, method=fdr_method)
    else:
        reject = pvals < alpha
        pval_corrected = pvals

    resp_corrected = [r[:,reject] for r in resp]

    stats = {'pval': pval, 'stat': statistics, 'significant': reject, 'alpha': alpha}

    if return_outstruct:
        outstruct_copy[resp_fieldname] = resp_corrected
        return outstruct_copy, stats

    return resp, stats


    











from collections.abc import Iterable, Sequence
import numpy as np
from mne import Info

STRICT_FIELDS_REQUIRED = set(['name','sound','soundf','resp','dataf'])

class OutStruct(Iterable):
    '''
    Class for storing electrode response data along with
    task- and electrode-related variables. Under the hood, it consists
    of a list of dictionaries where each dictionary contains all the data
    for one trial.

    Please see the :ref:`example notebooks <working with outstructs>` for more detailed
    tutorials which demonstrate using the OutStruct object in different types of analysis.
    
    Parameters
    ----------
    data : dict or list of dictionaries
        The Nth ictionary defines the Nth trial data, typically for the Nth stimulus.
        Each dictionary must contain the same keys if passed in a list of multiple trials.
    strict : bool, default=False
        If True, requires strict adherance to the following standards:
        1) Each trial must contain at least the following fields:
        ['name','sound','soundf','resp','dataf']
        2) Each trial must contain the exact same set of fields
    
    Attributes
    ----------
    fields : list of strings
        Field names in the data.
    mne_info : mne.Info instance
        Measurement info object containing things like electrode locations
        (only if OutStruct is created from reading a file format like BIDS).
    info : dict
        Extra info (not trial-specific) that a user wants to store
        using outstruct.set_info or outstruct.update_info

    
    '''
    def __init__(self, data, strict=False):
        
        if isinstance(data, dict):
            data = [data]
            self._data = data
        elif isinstance(data, list):
            self._data = data
        else:
            raise TypeError(f'Can only create OutStruct from a dict or a list '
                            f'of dicts, but found type {type(data)}')
        self._strict = strict
        self._validate_new_out_data(data, strict=strict)
        self._info = dict()
        self._mne_info = None

                
    def set_field(self, fielddata, fieldname):
        '''
        Parameters
        ----------
        fielddata : list
            List containing data to add to each trial for this field. Must 
            be same length as this object
        fieldname : string
            Name of field to add. If this field already exists in the OutStruct
            then the current field will be overwritten.
        '''
        if not isinstance(fielddata, list):
            raise TypeError(f'Input data must be a list, but found {type(fielddata)}')
        if len(fielddata) != len(self):
            raise Exception('Length of field is not equal to length of this OutStruct')
        for i in range(len(self.data)):
            self.data[i][fieldname] = fielddata[i]
            
    def get_field(self, fieldname):
        '''
        Return all trials for a single field.
        
        Parameters
        ----------
        fieldname : string
            Which field to get.
        Returns
        -------
        field : list
            List containing each trial's value for this field.
        '''
        try:
            return [tmp[fieldname] for tmp in self.data]
        except KeyError:
            raise KeyError(f'Invalid fieldname: {fieldname} not found in data.')
            
    def __getitem__(self, index):
        '''
        Get either a trial or a field using bracket indexing. See examples
        below for details.

        Parameters
        ----------
        index : int or string
            Which trial to get, or which field to get.

        Returns
        -------
        data : dict, list, or OutStruct
            If index is an integer, returns the corresponding trial as a dict. If index
            is a string, returns the corresponding field, and if it is a list of strings,
            returns those fields together in a new OutStruct object.

        Examples
        --------
        >>> # Get a specific trial based on its index, which returns a dict
        >>> from naplib import OutStruct
        >>> trial_data = [{'name': 'Zero', 'trial': 0, 'resp': [[0,1],[2,3]]},
        ...               {'name': 'One', 'trial': 1, 'resp': [[4,5],[6,7]]}]
        >>> out = OutStruct(trial_data, strict=False)
        >>> out[0]
        {'name': 'Zero', 'trial': 0, 'resp': [[0, 1], [2, 3]]}

        >>> # Get a slice of trials, which returns an OutStruct object
        >>> out[:2]
        OutStruct of 2 trials containing 3 fields
        [{"name": <class 'str'>, "trial": <class 'int'>, "resp": <class 'list'>}
        {"name": <class 'str'>, "trial": <class 'int'>, "resp": <class 'list'>}]

        >>> # Get a list of trial data from a single field
        >>> out['name']
        ['TrialZero', 'TrialOne']
        >>> out[0]
        {'name': 'TrialZero', 'trial': 0, 'resp': [[0, 1], [2, 3]]}

        >>> # Get multiple fields using a list of fieldnames, which returns an OutStruct containing that subset of fields
        >>> out[['resp','trial']]
        OutStruct of 2 trials containing 2 fields
        [{"resp": <class 'list'>, "trial": <class 'int'>}
        {"resp": <class 'list'>, "trial": <class 'int'>}]
        '''
        if isinstance(index, slice):
            return OutStruct(self.data[index], strict=self._strict)
        if isinstance(index, str):
            return self.get_field(index)
        if isinstance(index, list) or isinstance(index, np.ndarray):
            if isinstance(index[0], str):
                return OutStruct([dict([(field, x[field]) for field in index]) for x in self], strict=False)
            else:
                return OutStruct([self.data[i] for i in index], strict=False)
        try:
            # TODO: change this to return a type OutStruct if you do slicing - problem with trying to
            # print because it says KeyError for self.data[0] for key 0
            return self.data[index]
        except IndexError:
            raise IndexError(f'Index invalid for this data. Tried to index {index} but length is {len(self)}.')
        
            
    def __setitem__(self, index, data):
        '''
        Set a specific trial or set of trials, or set a specific field, using
        bracket indexing. See examples below for details.

        Parameters
        ----------
        index : int or string
            Which trial to set, or which field to set. If an integer, must be <= the
            length of the OutStruct, since you can only set a currently existing trial
            or append to the end, but you cannot set a trial that is beyond that. 
        data : dict or list of data
            Either trial data to add or field data to add. If index is an
            integer, dictionary should contain all the same fields as
            current OutStruct object.

        Examples
        --------
        >>> # Set a field of an OutStruct
        >>> from naplib import OutStruct
        >>> trial_data = [{'name': 'Zero', 'trial': 0, 'resp': [[0,1],[2,3]]},
        ...               {'name': 'One', 'trial': 1, 'resp': [[4,5],[6,7]]}]
        >>> out = OutStruct(trial_data)
        >>> out[0] = {'name': 'New', 'trial': 10, 'resp': [[0,-1],[-2,-3]]}
        >>> out[0]r 
        {'name': 'New', 'trial': 10, 'resp': [[0, -1], [-2, -3]]}

        >>> # We can also set all values of a field across trials
        >>> out['name'] = ['TrialZero','TrialOne']
        >>> out['name']
        ['TrialZero', 'TrialOne']
        '''
        if isinstance(index, str):
            self.set_field(data, index)
        else:
            if index > len(self):
                raise IndexError((f'Index is too large. Current data is length {len(self)} '
                    'but tried to set index {index}. If you want to add to the end of the list '
                    'of trials, use the OutStruct.append() method.'))
            elif index == len(self):
                self.append(data)
            else:
                self.data[index] = data
     
    def append(self, trial_data, strict=None):
        '''
        Append a single trial of data to the end of an OutStruct.
        
        Parameters
        ----------
        trial_data : dict
            Dictionary containing all the same fields as current OutStruct object.
        strict : bool, default=self._strict
            If true, enforces that new data contains the exact same set of fields as
            the current OutStruct. Default value is self._strict, which is set based
            on the input when creating a new OutStruct from scratch with __init__()

        Raises
        ------
        TypeError
            If input data is not a dict.
        ValueError
            If strict is `True` and the fields contained in the trial_data do
            not match the fields currently stored in the OutStruct.

        Examples
        --------
        >>> # Set a field of an OutStruct
        >>> from naplib import OutStruct
        >>> trial_data = [{'name': 'Zero', 'trial': 0, 'resp': [[0,1],[2,3]]},
        ...               {'name': 'One', 'trial': 1, 'resp': [[4,5],[6,7]]}]
        >>> out = OutStruct(trial_data)
        >>> new_trial_data = {'name': 'Two', 'trial': 2, 'resp': [[8,9],[10,11]]}
        >>> out.append(new_trial_data)
        >>> len(out)
        3
        '''
        if strict is None:
            strict = self._strict
        self._validate_new_out_data([trial_data], strict=strict)
        self.data.append(trial_data)
        
    def set_info(self, info):
        '''
        Set the info dict for this OutStruct. If there is already data in the
        `info` attribute, it is replaced with this.
        
        Parameters
        ----------
        info : dict
            Dictionary containing info to store in the OutStruct's `info` attribute.
            
        '''
        if not isinstance(info, dict):
            raise TypeError(f'info must be a dict but got {type(info)}')
        self._info = info
        
    def update_info(self, info):
        '''
        Add data from a dict to this object's `info` attribute. If there is already data in the
        `info` attribute, this new info is simply added. Keys which exist in the current
        `info` dict and also in this new dict will be replaced, while others will be kept.
        
        Parameters
        ----------
        info : dict
            Dictionary containing info to add to the OutStruct's `info` attribute.
        '''
        self._info.update(info)
        
    def set_mne_info(self, info):
        '''
        Set the mne_info attribute, which contains measurement information.
        
        Parameters
        ----------
        info : mne.Info instance
            Info to set.
        '''
        if not isinstance(info, Info):
            raise TypeError(f'input info must be an instance of mne.Info, but got {type(info)}')
        self._mne_info = info
    
    def __iter__(self):
        return (self[i] for i in range(len(self)))

    def __len__(self):
        '''
        Get the number of trials in the OutStruct with ``len(outstruct)``.

        Examples
        --------
        >>> from naplib import OutStruct
        >>> trial_data = [{'trial': 0, 'resp': [[0,1],[2,3]]},
                          {'trial': 1, 'resp': [[4,5],[6,7]]}]
        >>> out = OutStruct(trial_data, strict=False)
        >>> len(out)
        2
        '''
        return len(self.data)
    
    def __repr__(self):
        return self.__str__() # until we can think of a better __repr__
    
    def __str__(self):
        to_return = f'OutStruct of {len(self)} trials containing {len(self.fields)} fields\n['
        
        to_print = 2 if len(self) > 3 else 3
        for trial_idx, trial in enumerate(self[:to_print]):
            fieldnames = list(trial.keys())
            to_return += '{'
            for f, fieldname in enumerate(fieldnames):
#                 to_return += f'"{fieldname}": {trial[fieldname].__str__()}'
                to_return += f'"{fieldname}": {type(trial[fieldname])}'
                if f < len(fieldnames)-1:
                    to_return += ', '
            if trial_idx < len(self)-1:
                to_return += '}\n'
            else:
                to_return += '}'
        if to_print == 3:
             to_return += ']\n'
        elif to_print == 2:
            to_return += '\n...\n{'
            fieldnames = list(self[-1].keys())
            for f, fieldname in enumerate(fieldnames):
#                 to_return += f'"{fieldname}": {self[-1][fieldname].__str__()}'
                to_return += f'"{fieldname}": {type(self[-1][fieldname])}'
                if f < len(fieldnames)-1:
                    to_return += ', '
            to_return += '}]'
        return to_return
    
    def _validate_new_out_data(self, input_data, strict=True):
        
        first_trial_fields = set(self.fields)
        for t, trial in enumerate(input_data):
            if not isinstance(trial, dict):
                raise TypeError(f'input data is not a list of dicts, found {type(trial)}')
            trial_fields = set(trial.keys())
            if strict and trial_fields != first_trial_fields:
                raise ValueError(f'New data does not contain the same fields as the first trial.')        
            if strict:
                for required_field in STRICT_FIELDS_REQUIRED:
                    if required_field not in trial_fields:
                        raise ValueError(f'For a "strict" OutStruct, the data does not contain the required field {required_field}.')
    
    @property
    def fields(self):
        '''List of strings containing names of all fields in this OutStruct.'''
        return [k for k, _ in self._data[0].items()]
    
    @property
    def data(self):
        '''List of dictionaries containing data for each stimulus
        response and all associated variables.'''
        return self._data

    @property
    def info(self):
        '''Dictionary which can be used to store metadata info which does not
        change over trials, such as subject, recording, or task information.'''
        return self._info
    
    @property
    def mne_info(self):
        '''
        `mne.Info <https://mne.tools/dev/generated/mne.Info.html>`_ instance
        which stores measurement information and can be used with mne's visualization
        functions. This is empty by default unless it is manually added or read in
        by a function like `naplib.io.read_bids`.
        '''
        if self._mne_info is None:
            raise ValueError('No mne_info is available for this OutStruct. This must '
                             'be read in from external data or added manually to the OutStruct.')
        return self._mne_info
    
def join_fields(outstructs, fieldname='resp', axis=-1, return_outstruct=False):
    '''
    Join trials from a field of multiple OutStruct objects by zipping them
    together and concatenating each trial together. The field must be of type
    np.ndarray and concatenation is done with np.concatenate().
    
    Parameters
    ----------
    outstructs : sequence of OutStructs
        Sequence containing the different outstructs to join
    fieldname : string, default='resp'
        Name of the field to concatenate from each OutStruct. For each trial in
        each outstruct, this field must be of type np.ndarray or something which
        can be input to np.concatenate().
    axis : int, default = -1
        Axis along which to concatenate each trial's data. The default corresponds
        to the channel dimension of the conventional 'resp' field of an OutStruct.
    return_outstruct : bool, default=False
        If True, returns data as an OutStruct with a single field named fieldname.

    Returns
    -------
    joined_data : list of np.ndarrays, or OutStruct
        Joined data of same length as each of the outstructs containing concatenated data
        for each trial.
    '''
    
    for out in outstructs:
        if not isinstance(out, OutStruct):
            raise TypeError(f'All inputs must be an OutStruct but found {type(out)}')
        field = out.get_field(fieldname)
        if not isinstance(field[0], np.ndarray):
            raise TypeError(f'Can only concatenate np.ndarrays, but found {type(field[0])} in this field')

    starting_fields = [out.get_field(fieldname) for out in outstructs] # each one should be a list of np.arrays
    
    to_return = []
    
    zipped_fields = list(zip(*starting_fields))
    for i, field_set in enumerate(zipped_fields):
        to_return.append(np.concatenate(field_set, axis=axis))
        
    if return_outstruct:
        return OutStruct([dict([(fieldname, x)]) for x in to_return], strict=False)
    return to_return
        
    



# Copyright (c) 2011-2014 Kyle Gorman and Michael Wagner
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Corpus utilities
"""


import os
import logging

from re import match
from glob import glob
from shutil import rmtree
from tempfile import mkdtemp
from subprocess import check_call

from wavfile import WavFile
from prondict import PronDict
from utilities import splitname, mkdir_p, opts2cfg, \
                       MISSING, OOV, SIL, SP, TEMP


# regexp for inspecting phones
VALID_PHONE = r"^[^\d\s]+[0-9]?$"


class Corpus(object):

    """
    Class representing directory of training data; once constructed, it
    is ready for training or aligning.
    """

    def __init__(self, dirname, opts):
        # temporary directories for stashing the data
        tmpdir = os.environ["TMPDIR"] if "TMPDIR" in os.environ else None
        self.tmpdir = mkdtemp(dir=tmpdir)
        self.auddir = os.path.join(self.tmpdir, "audio")
        mkdir_p(self.auddir)
        self.labdir = os.path.join(self.tmpdir, "label")
        mkdir_p(self.labdir)
        # samplerate
        self.samplerate = opts["samplerate"]
        # phoneset
        self.phoneset = frozenset(opts["phoneset"])
        for phone in self.phoneset:
            if not match(VALID_PHONE, phone):
                logging.error("Phone '{}': not /{}/.".format(phone,
                                                      VALID_PHONE))
                exit(1)
        # dictionaries
        self.dictionary = []
        if "dictionary" in opts and opts["dictionary"]:
            assert type(opts["dictionary"]) is list
            self.dictionary.extend(opts["dictionary"])
        self.thedict = PronDict(self.phoneset)
        for dic in self.dictionary:
            self.thedict.add(dic)
        #self.thedict[SIL] = [SIL]
        self.taskdict = os.path.join(self.tmpdir, "taskdict")
        # word and phone lists
        self.phons = os.path.join(self.tmpdir, "phons")
        self.words = os.path.join(self.tmpdir, "words")
        # MLF
        self.pron_mlf = os.path.join(self.tmpdir, "pron.mlf")
        self.word_mlf = os.path.join(self.tmpdir, "words.mlf")
        self.phon_mlf = os.path.join(self.tmpdir, "phones.mlf")
        # feature extraction configuration
        self.HCopy_cfg = os.path.join(self.tmpdir, "HCopy.cfg")
        opts2cfg(self.HCopy_cfg, opts["HCopy"])
        self.audio_scp = os.path.join(self.tmpdir, "audio.scp")
        self.feature_scp = os.path.join(self.tmpdir, "feature.scp")
        # prepare the data for processing
        (audiofiles, labelfiles) = self._lists(dirname)
        self.audiofiles = audiofiles
        self._prepare_label(labelfiles)
        self._prepare_audio(audiofiles)
        self._extract_features()

    def _lists(self, dirname):
        """
        Create lists of .wav and .lab files, detecting missing pairs.
        """
        audiofiles = glob(os.path.join(dirname, "*.wav"))
        labelfiles = glob(os.path.join(dirname, "*.lab"))
        if not audiofiles:
            logging.error("No .wav files in '{}'.".format(dirname))
            exit(1)
        elif not labelfiles:
            logging.error("No .lab files in '{}'.".format(dirname))
            exit(1)
        audiobasenames = frozenset(splitname(audiofile)[1] for
                                   audiofile in audiofiles)
        labelbasenames = frozenset(splitname(labelfile)[1] for
                                   labelfile in labelfiles)
        missing = []
        missing.extend(basename + ".wav" for basename in
                       labelbasenames - audiobasenames)
        missing.extend(basename + ".lab" for basename in
                       audiobasenames - labelbasenames)
        if missing:
            with open(MISSING, "w") as sink:
                for filename in missing:
                    print(os.path.join(dirname, filename), file=sink)
            logging.error("Missing data files: see '{}'.".format(MISSING))
            exit(1)
        return (audiofiles, labelfiles)

    def _prepare_label(self, labelfiles):
        """
        Check label files against dictionary, and construct new .lab
        and .mlf files
        """
        found_words = set()
        with open(self.word_mlf, "w") as word_mlf:
            print("#!MLF!#", file=word_mlf)
            for labelfile in labelfiles:
                (_, filename) = os.path.split(labelfile)
                phon_labfile = os.path.join(self.auddir, filename)
                word_labfile = os.path.join(self.labdir, filename)
                # header for each file in the .mlf
                print('"{}"'.format(word_labfile), file=word_mlf)
                # read in words from original .lab file
                with open(labelfile, "r") as orig_handle:
                    words = orig_handle.readline().split()
                found_words.update(words)
                # write out new wordlab
                with open(word_labfile, "w") as word_handle:
                    print("\n".join(words), file=word_handle)
                # get pronunciation and check for in-dictionary-hood
                phons = []
                for word in words:
                    try:
                        phons.extend(self.thedict[word][0])
                    except (KeyError, IndexError):
                        pass
                # write out new phonelab
                with open(phon_labfile, "w") as phon_handle:
                    print("\n".join(phons), file=phon_handle)
                # append to word_mlf
                print("\n".join(words), file=word_mlf)
                print(".", file=word_mlf)
        # report and die if OOV words are found
        if self.thedict.oov:
            with open(OOV, "w") as oov:
                print("\n".join(sorted(self.thedict.oov)), file=oov)
            logging.error("OOV word(s): see '{}'.".format(OOV))
            exit(1)
        # make words
        with open(self.words, "w") as words:
            print("\n".join(found_words), file=words)
        # create temp file to abuse
        temp = os.path.join(self.tmpdir, TEMP)
        # run HDMan
        with open(temp, "w") as ded:
            print("""AS {0}\nMP {1} {1} {0}
""".format(SP, SIL), file=ded)
        check_call(["HDMan", "-m",
                             "-g", temp,
                             "-w", self.words,
                             "-n", self.phons,
                             self.taskdict] +
                   self.dictionary)
        # add SIL to phone list
        with open(self.phons, "a") as phons:
            print(SIL, file=phons)
        # add SIL to taskdict
        with open(self.taskdict, "a") as taskdict:
            print("{0} {0}".format(SIL), file=taskdict)
        # run HLEd
        with open(temp, "w") as led:
            print("""EX
IS {0} {0}
DE {1}
""".format(SIL, SP), file=led)
        check_call(["HLEd", "-l", self.labdir,
                            "-d", self.taskdict,
                            "-i", self.phon_mlf,
                            temp, self.word_mlf])

    def _prepare_audio(self, audiofiles):
        """
        Check audio files, downsampling if necessary, creating .scp file
        """
        with open(self.audio_scp, "w") as audio_scp, \
                open(self.feature_scp, "w") as feature_scp:
            for audiofile in audiofiles:
                (_, filename) = os.path.split(audiofile)
                (basename, _) = os.path.splitext(filename)
                featurefile = os.path.join(self.auddir, basename + ".mfc")
                Fs = WavFile.samplerate(audiofile)
                if Fs != self.samplerate:
                    w = WavFile.from_file(audiofile)
                    new_wav = os.path.join(self.auddir, filename)
                    logging.warning("Resampling '{}'.".format(audiofile))
                    w.resample_bang(self.samplerate)
                    w.write(new_wav)
                print('"{}" "{}"'.format(audiofile, featurefile),
                      file=audio_scp)
                print('"{}"'.format(featurefile), file=feature_scp)

    def _extract_features(self):
        """
        Compute audio features
        """
        check_call(["HCopy", "-C", self.HCopy_cfg, "-S", self.audio_scp])

    def __del__(self):
        rmtree(self.tmpdir) 



import naplib.features
import naplib.segmentation
import naplib.stats
import naplib.visualization
import naplib.encoding
import naplib.io
import naplib.preprocessing
import naplib.array_ops
import naplib.model_selection
import naplib.utils
from .out_struct import OutStruct, join_fields

__version__ = "0.1.0"



from itertools import groupby

from ..out_struct import OutStruct

def _parse_outstruct_args(outstruct,
                          *args,
                          allow_different_lengths=False,
                          allow_strings_without_outstruct=True):
    '''
    Parse a variable number of arguments and return a tuple with either the field
    from the outstruct which is specified by a given arg or the arg itself if it
    does not correspond to a field in the outstruct.
    
    Parameters
    ----------
    outstruct : any type or None
        If a naplib.OutStruct object, then will check if the other args are fields
        of this object, otherwise they will be returned without change.
    *args : variable number of inputs
        Other arguments. Each one will be checked to see if it is a field of the
        outstruct, and if it is then it will be replaced by that field. If it is a
        list, it will be returned without change, and if it is any other type, it will
        be repeated into a list of the same length as the outstruct or any other list args,
        and if nothing else is provided, it will be return as is.
    allow_different_lengths : bool, default=False
        If True, args are not required to all have the same length, and they do not need
        to be the same length as the outstruct if one is provided.
    allow_strings_without_outstruct : bool | list of bool (length len(args)), default=True
        If True, then string args that are passed in are accepted and returned back
        even if no outstruct was provided. 

    Returns
    -------
    *args_out : single arg or tuple of args, same length as *args input
    '''
    # if no outstruct was provided, then just parse the args on their own
    if not isinstance(outstruct, OutStruct):
        args = list(args)
        if isinstance(allow_strings_without_outstruct, bool):
            allow_strings_without_outstruct = [allow_strings_without_outstruct for _ in range(len(args))]
        if len(args) > 1:
            lengths = []
            for i, arg in enumerate(args):
                if isinstance(arg, list):
                    lengths.append(len(arg))
                if not allow_strings_without_outstruct[i]:
                    raise ValueError(f'Input is a string but no OutStruct was provided from which to select a field based on the string.')
            if len(lengths) > 0:
                if not allow_different_lengths and not _all_equal_list(lengths):
                    raise ValueError(f"The supplied list arguments have different lengths, but they must all be the same length.")
                for i, arg in enumerate(args):
                    if not isinstance(arg, list) and not isinstance(arg, tuple):
                        args[i] = [arg for _ in range(lengths[0])]
            return tuple(args)
        else:
            return args[0]
    
    # if an outstruct was provided, check if args are fields of the outstruct
    args_out = []
    
    outstruct_fieldnames = outstruct.fields
    for arg in args:
        if isinstance(arg, str):
            if arg in outstruct_fieldnames:
                args_out.append(outstruct.get_field(arg))
            else:
                raise ValueError(f"{arg} is a string but is not a field of the outstruct.")
        else:
            if isinstance(arg, list):
                if not allow_different_lengths and len(arg) != len(outstruct):
                    raise ValueError(f"The list argument supplied is length {len(arg)}, which is not"
                                     " the same as the provided OutStruct, which is length {len(outstruct)}.")
                args_out.append(arg)
            elif isinstance(arg, tuple):
                args_out.append(arg)
            else:
                args_out.append([arg for _ in range(len(outstruct))])
    if len(args_out) > 1:
        return tuple(args_out)
    return args_out[0]
                            
def _all_equal_list(iterable):
    g = groupby(iterable)
    return next(g, True) and not next(g, False)




#!/usr/bin/env python3
# Kyle Gorman <gormanky@ohsu.edu>
# The UNIX `sort` utility does not always sort the dictionary the way that
# HTK expects; this one does


import fileinput


if __name__ == "__main__":
    lines = frozenset(l.rstrip() for l in fileinput.input())  # accumulate
    print("\n".join(sorted(lines)))                           # linearize



from itertools import chain

import numpy as np
from sklearn.model_selection._split import _BaseKFold
from sklearn.utils.validation import indexable
from sklearn.utils import _safe_indexing, check_random_state

from ..out_struct import OutStruct


class KFold(_BaseKFold):
    '''
    KFold splitter which works on an OutStruct or a list.
    
    Parameters
    ----------
    n_splits : int
        Number of folds. Must be at least 2.
    shuffle : bool, default=False
        Whether to shuffle the data before splitting into batches.
        Note that the samples within each split will not be shuffled.
    random_state : int, RandomState instance or None, default=None
        When `shuffle` is True, `random_state` affects the ordering of the
        indices, which controls the randomness of each fold. Otherwise, this
        parameter has no effect.
        Pass an int for reproducible output across multiple function calls.
        
    Examples
    --------
    >>> from naplib.model_selection import KFold
    >>> data = [1,2,3] # this could be a field of an OutStruct, like out['resp']
    >>> data2 = [5,6,7] # this could be another field of an OutStruct, like out['aud']
    >>> kfold = KFold(3)
    >>> for train_data, test_data, train_data2, test_data2 in kfold.split(data, data2):
    >>>    print(train_data, test_data, train_data2, test_data2)
    [2, 3] [1] [6, 7] [5]
    [1, 3] [2] [5, 7] [6]
    [1, 2] [3] [5, 6] [7]
    
    '''
    def __init__(self, n_splits, shuffle=False, random_state=None):
         super().__init__(n_splits=n_splits, shuffle=shuffle, random_state=random_state)
            
    def _iter_test_indices(self, X, y=None, groups=None):
        n_samples = len(X)
        indices = np.arange(n_samples)
        if self.shuffle:
            check_random_state(self.random_state).shuffle(indices)

        n_splits = self.n_splits
        fold_sizes = np.full(n_splits, n_samples // n_splits, dtype=int)
        fold_sizes[: n_samples % n_splits] += 1
        current = 0
        for fold_size in fold_sizes:
            start, stop = current, current + fold_size
            yield indices[start:stop]
            current = stop

    def split(self, *args):
        """Generate splits of the data.

        Parameters
        ----------
        *args : OutStructs or lists
            Sets of data which will be split into train and test groups.

        Yields
        ------
        train : OutStruct or list
            The training set for that split.
        test : OutStruct or list
            The testing set for that split.
        """
        data = indexable(*args)
        n_samples = len(data[0])
        if self.n_splits > n_samples:
            raise ValueError(
                (
                    "Cannot have number of splits n_splits={0} greater"
                    " than the number of samples: n_samples={1}."
                ).format(self.n_splits, n_samples)
            )

        for train, test in super().split(data[0]):
            tmp = list(
                    chain.from_iterable(
                        (_safe_indexing(a, train), _safe_indexing(a, test)) for a in data
                    )
                )
            for i, d in enumerate(data):
                if isinstance(d, OutStruct):
                    tmp[2*i] = OutStruct(tmp[2*i], strict=False)
                    tmp[2*i+1] = OutStruct(tmp[2*i+1], strict=False)
            yield tmp
                    
    




import numpy as np

def dprime(D, L):
    '''
    Compute DPrime.
    
    Parameters
    ----------
    D : array-like of shape (features, instance)
        Data features.
    L : array-like of shape (instance, )
        Labels for each instance.ß
    
    Returns
    -------
    allf : np.ndarray
        Allf
    f_stat : float
        F-statistic.
    f_std : float
        Standard dev of allf.

    '''
    
    labels = np.unique(L).squeeze()
        
    Dtot = []
    Ltot = []
    
    for i in range(len(labels)):
        Dtot.append(D[:, np.argwhere(L==labels[i])])
        Ltot.append(L[np.argwhere(L==labels[i])])
        
    D = np.concatenate(Dtot, axis=1).squeeze(-1)
    L = np.concatenate(Ltot, axis=0).squeeze()
    
    
    xh = D.mean(-1).squeeze()
    total = 0.0
        
    
    sbs = []
    sws = []
    
    for cnt1, cnt2 in enumerate(labels):
        # first, between class variability
        index = np.argwhere(L==cnt2)
        index = index.flatten()
        dc = D[:, index].mean(1)
        sb = np.linalg.norm(dc - xh, 2)**2
        
        # now, within class variability
        sw = 0.0
        for cnt3 in index:
            sw += np.linalg.norm(D[:, cnt3] - dc, 2)**2
        
        sbs.append(sb * len(index))
        sws.append(sw)
        total += len(index)
        
    sbs = np.array(sbs)
    sws = np.array(sws)
        
    if np.min(sws) == 0:
        sws[sws==0] = 1e-6
    
    tmpb = sbs / (len(labels)-1)
    tmpw = sws / (total - len(labels))
    allf = np.divide(tmpb, tmpw)
    
    f = tmpb.sum() / tmpw.sum()
    
    fs = np.std(len(labels)*tmpb / tmpw.sum(), ddof=1) / np.sqrt(len(labels))
    
    return allf, f, fs
        
def fratio(D, L, elec_mode='all'):
    '''
    F ratio over time.
    
    Parameters
    ----------
    D : array-like of shape (electrodes, time, instances)
        Data features over time.
    L : array-like containing labels for each instance
        L is either of shape (1, instances) if each instance has the same 
        label across the full time axis, or is of shape (time, instances)
    elec_mode : string, one of ['all', 'individual']
        if 'all', computes f-ratio over all electrodes together, and returns
        f-ratio of shape (time,)
        if 'individual', computes f-ratio for each electrode individually
        and returns f-ratio of shape (electrodes, time)
        
    Returns
    -------
    fratio : np.ndarray
        if elec_mode=='all', shape=(time,)
        if elec_mode=='individual', shape=(electrodes, time)
    
    '''
    
    ndim_L = L.squeeze().ndim
    
    label = np.unique(L)
    
    f = []
    if elec_mode == 'all':
        f = np.zeros((D.shape[1]))
        for t in range(D.shape[1]):
            tmp = D[:,t,:].squeeze()
            if ndim_L > 1:
                _, f_tmp, _ = dprime(tmp, L[t,:].squeeze())
            else:
                _, f_tmp, _ = dprime(tmp.reshape(tmp.shape[0], -1), L.squeeze())
            f[t] = f_tmp
        
    elif elec_mode == 'individual':
        f = np.zeros((D.shape[0], D.shape[1]))
        for t in range(D.shape[1]):
            tmp = D[:,t,:].squeeze()
            for cnt in range(D.shape[0]):
                if ndim_L > 1:
                    _, f_tmp, _ = dprime(tmp[cnt,:], L[t,:].squeeze())
                else:
                    _, f_tmp, _ = dprime(tmp[cnt,:].reshape(1,-1), L.squeeze())
                f[cnt,t] = f_tmp
        
    else:
        raise Exception(f'Error: elec_mode should be one of ["all", "individual"], but got {elec_mode}')
        
    return f
    
    



# Copyright (c) 2011-2014 Kyle Gorman and Michael Wagner
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Pronunciation dictionary utilities
"""


import logging

from collections import defaultdict

from utilities import SIL, SP


class PronDict(object):

    """
    A wrapper for a normal pronunciation dictionary in the CMU style
    """

    SILENT_PHONES = frozenset([SIL, SP])

    @staticmethod
    def pronify(source):
        for (i, line) in enumerate(source, 1):
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            try:
                (word, pron) = line.split(None, 1)
            except ValueError:
                logging.error("Formatting error in dictionary '{}' (ln. {}).".format(source.name, i))
                exit(1)
            yield (i, word, pron.split())

    def add(self, filename):
        # build up dictionary
        with open(filename, "r") as source:
            for (i, word, pron) in PronDict.pronify(source):
                for ph in pron:
                    if ph not in self.ps | self.SILENT_PHONES:
                        logging.error("Unknown phone '{}' in dictionary '{}' (ln. {}).".format(ph, filename, i))
                        exit(1)
                self.d[word].append(pron)

    def __init__(self, phoneset, filename=None):
        self.ps = phoneset
        self.d = defaultdict(list)
        if filename:
            self.add(filename)
        # for later...
        self.oov = set()

    def __contains__(self, key):
        return key in self.d and self.d[key] != []

    def __getitem__(self, key):
        getlist = self.d[key]
        if getlist:
            return getlist
        else:
            self.oov.add(key)
            raise KeyError(key)

    def __repr__(self):
        return "PronDict({})".format(self.d)

    def __setitem__(self, key, value):
        self.d[key].append(value)



from copy import deepcopy
import numpy as np

def _extract_windows_vectorized(arr, clearing_time_index, max_time, sub_window_size):
    '''
    Vectorized method to extract sub-windows of an array.
    '''
    start = clearing_time_index + 1 - sub_window_size + 1
    
    sub_windows = (
        start +
        # expand_dims are used to convert a 1D array to 2D array.
        np.expand_dims(np.arange(sub_window_size), 0) +
        np.expand_dims(np.arange(max_time + 1), 0).T
    )
    return arr[sub_windows]

def sliding_window(arr, window_len, window_key_idx=0, fill_out_of_bounds=True, fill_value=0):
    '''
    Extract windows of length window_len and put them into an array. Can be
    used for causal, anticausal, or noncausal windowing.
    
    Parameters
    ----------
    arr : np.ndarray, shape (time, *feature_dims)
        Data to be windowed. Windowing is only applied across first dimension,
        which is assumed to be time. All other dimensions are kept the same for
        the output.
    
    window_len : int
        length of sliding window
        
    window_key_idx : int, default=0 (must be from 0 to window_len-1)
        Key point of a given sliding window. A value of 0 corresponds to causal sliding
        windows, where the first window_len-1 values in the nth window
        happen before the nth point in arr. A value of window_len corresponds to
        anti-causal sliding windows, where the first value in the nth window is
        arr[n], and the remaining window_len-1 values come after that point. A value
        of 1 would return windows where the nth window is a window starting at
        arr[n-(window_len-2)] and ending at (and including) arr[n+1].
    
    fill_out_of_bounds : bool, default=True
        If True, prepends fill_value to the first (window_len-1) samples before
        the beginning of the array across all feature dimensions
        so that the output is the same length as the input 
        (i.e. there is one window for each time point in the
        original array, though the first window will contain only zeros except
        for the last value). If False, does not prepend zeros, so the output
        has fewer windows than the input has time points.
    
    Returns
    -------
    windows : np.ndarray, shape (n_samples, window_len, *feature_dims...)
        Windowed array segments.
    
    
    Examples
    --------
    >>> import numpy as np
    >>> from naplib.array_ops import sliding_window
    >>> arr = np.arange(1,5)
    >>> slide1 = sliding_window(arr, 3)
    >>> slide2 = sliding_window(arr, 3, 0, False)
    >>> slide3 = sliding_window(arr, 3, 2)
    >>> slide4 = sliding_window(arr, 3, 1)
    >>> print(slide1)
    [[0. 0. 1.]
     [0. 1. 2.]
     [1. 2. 3.]
     [2. 3. 4.]]
    >>> print(slide2)
    [[1 2 3]
     [2 3 4]]
    >>> print(slide3)
    [[1. 2. 3.]
     [2. 3. 4.]
     [3. 4. 0.]
     [4. 0. 0.]]
    >>> print(slide4)
    [[0. 1. 2.]
     [1. 2. 3.]
     [2. 3. 4.]
     [3. 4. 0.]]
    '''
    
    if fill_out_of_bounds:
        if window_key_idx == 0:
            arr = np.concatenate([fill_value*np.ones([window_len-1-window_key_idx, *arr.shape[1:]]), arr], axis=0)
        elif window_key_idx == window_len - 1:
            arr = np.concatenate([arr, fill_value*np.ones([window_key_idx, *arr.shape[1:]])], axis=0)
        elif window_key_idx < window_len - 1:
            arr = np.concatenate([fill_value*np.ones([window_len-1-window_key_idx, *arr.shape[1:]]), arr, fill_value*np.ones([window_key_idx, *arr.shape[1:]])], axis=0)
        else:
            raise ValueError(f'window_key_idx must be an integer from 0 to window_len-1, but got {window_key_idx}')
    
    return _extract_windows_vectorized(arr, window_len-2, arr.shape[0]-window_len, window_len)

def concat_apply(data_list, function, axis=0, function_kwargs=None):
    '''
    Apply a function to a list of data by first contatenating the
    list into a single array along the `axis` dimension, passing it into the function,
    and then spreading the result back into the same size list.
    The function must return an array with the `axis` dimension unchanged.
    
    Parameters
    ----------
    data_list : list of np.array's
        Each array in the list must match in all dimensions except for `axis` so
        that they can be concatenated along that dimension.
        
    function : Callable
        A function which operates on an array. It must return an array where the
        `axis` dimensions is unchanged. For example, this could be something like 
        sklearn.manifold.TSNE().fit_transform if `axis=0`, or your own custom function.
    
    axis : int, default=0
        Axis over which to concatenate and then re-split the data_list before
        and after applying the function.

    function_kwargs : dict, default=None
        If provided, a dict of keyword arguments to pass to the function.

    Returns
    -------
    output : list of np.ndarray's
        List of arrays after chopping up the output of the function into arrays
        of the same length as the original input.
    
    Raises
    ------
    RuntimeError: If the callable function changes the size of the concatenation/splitting axis.
    
    Examples
    --------
    >>> import numpy as np
    >>> from naplib.array_ops import concat_apply
    >>> data = [np.arange(20).reshape((5,4)), np.arange(20, 40).reshape((5,4))] # 2 trials, 5 samples with 4 channels
    >>> data
    [array([[ 0,  1,  2,  3],
            [ 4,  5,  6,  7],
            [ 8,  9, 10, 11],
            [12, 13, 14, 15],
            [16, 17, 18, 19]]),
     array([[20, 21, 22, 23],
            [24, 25, 26, 27],
            [28, 29, 30, 31],
            [32, 33, 34, 35],
            [36, 37, 38, 39]])]

    >>> # We can use PCA to reduce the channel dimensionality by fitting PCA on the
    >>> # concatenated data, transforming it, and then splitting it back into 2 trials
    >>> from sklearn.decomposition import PCA
    >>> data_pca = concat_apply(data, PCA(2).fit_transform)
    >>> data_pca
    [array([[-3.60000000e+01,  8.63623587e-15],
            [-2.80000000e+01, -2.36903429e-15],
        [-2.00000000e+01, -1.34899193e-15],
            [-1.20000000e+01, -5.15542367e-16],
            [-4.00000000e+00, -4.16724783e-16]]),
     array([[4.00000000e+00, 4.16724783e-16],
            [1.20000000e+01, 5.15542367e-16],
            [2.00000000e+01, 1.34899193e-15],
            [2.80000000e+01, 2.36903429e-15],
            [3.60000000e+01, 3.01589107e-15]])]

    >>> # We can downsample the channel dimension, making use of
    >>> # the function_kwargs parameter
    >>> from scipy.signal import resample
    >>> downsampled_channels = concat_apply(data, resample, function_kwargs={'num': 3, 'axis': 1})
    >>> downsampled_channels
    [array([[ 0.5      ,  1.1339746,  2.8660254],
            [ 4.5      ,  5.1339746,  6.8660254],
            [ 8.5      ,  9.1339746, 10.8660254],
            [12.5      , 13.1339746, 14.8660254],
            [16.5      , 17.1339746, 18.8660254]]),
     array([[20.5      , 21.1339746, 22.8660254],
            [24.5      , 25.1339746, 26.8660254],
            [28.5      , 29.1339746, 30.8660254],
            [32.5      , 33.1339746, 34.8660254],
            [36.5      , 37.1339746, 38.8660254]])]
    '''
    lengths = np.array([x.shape[axis] for x in data_list])
    data_cat = np.concatenate(data_list, axis=axis)
    
    N = data_cat.shape[axis]
    
    if function_kwargs is None:
        function_kwargs = {}
    if not isinstance(function_kwargs, dict):
        raise TypeError(f'function_kwargs must be a dict of keyword arguments, but got {type(function_kwargs)}')

    func_output = function(data_cat, **function_kwargs)
    
    if func_output.shape[axis] != N:
        raise RuntimeError(f'The callable function changed the size of the concatenation and splitting axis from {N} to {func_output.shape[axis]}, but this dimension must remain constant.')

    # split output back into list, but cut off the last because it is an empty array
    output = [x for x in np.split(func_output, np.cumsum(lengths), axis=axis)[:-1]]
    
    return output



from os.path import dirname, join
import math
import numpy as np
from scipy.signal import resample, lfilter
from hdf5storage import loadmat

def _largest_pow2_less_than(x):
    '''
    Get largest power of 2 less than x.
    '''
    return 2**(math.ceil(math.log(x, 2)) - 1)

def _sigmoid(x, factor):
    '''
    Nonlinear sigmoid for cochlear model which simulates hair
    cell nonlinearity.
    
    Parameters
    ----------
    x : float
        Data to pass through nonlinear function
    factor: float
        Nonlinear factor.
        If >0, transistor-like function. 
        If =0, hard-limiter
        If =-1, half-wave rectifier
        Else, no operation and x is returned unchanged, i.e. linear
    '''
    if isinstance(factor, float):
        return 1 / (1 + np.exp(-x / factor))
    if factor == 'boolean':
        return (x>0).astype('float')
    elif factor == 'half-wave':
        return np.maximum(x, 0)
    else:
        return x
        

def auditory_spectrogram(x, sfreq, frame_len=8, tc=4, factor='linear'):
    '''
    Compute the auditory spectrogram of a signal using a model of the
    peripheral auditory system [#1yang]_. The model includes a cochlear
    filter bank of 128 constant-Q filters logarithmically-spaced, a hair
    cell model which includes a low-pass filter and a nonlinear compression
    function, and a lateral inhibitory network over the spectral axis. The
    envelope of each frequency band gives the time-frequency representation.
    
    Parameters
    ----------
    x : np.ndarray
        Acoustic signal to convert to time-frequency representation
    sfreq : int
        Sampling rate of the signal. If not a power of 2, then it will be
        downsampled to the nearest power of 2.
    frame_len : float, default=8
        Frame length of the output, in ms. Typically 8, 16, or a power of 2.
    tc : float, default=4
        Time constant for leaky integration. Must be >=0. If 0, then leaky
        integration becomes short-term average. Typically 4, 16, or 64 ms.
    factor : float or string, default='linear'
        Nonlinear factor for hair cell model. If a positive float, specifies the
        critical level factor (typically 0.1 for a unit sequence). The smaller the
        value, the more the compression. Or, can be one of
        {'linear', 'boolean', 'half-wave'} describing the compressor.
    
    Returns
    -------
    aud : np.ndarray
        Auditory spectrogram, shape (n_frames, 128).
    
    Notes
    -----
    This is a python re-implementation of the Matlab function `wav2aud` in the
    `NSLtools toolbox <https://github.com/tel/NSLtools>`_.
    
    References
    ----------
    .. [#1yang] X. Yang, K. Wang and S. A. Shamma, "Auditory representations of
        acoustic signals," in IEEE Transactions on Information Theory, vol.
        38, no. 2, pp. 824-839, March 1992, doi: 10.1109/18.119739.
    
    '''
    
    if frame_len <= 0:
        raise ValueError(f'frame_len must be positive but got {frame_len}')
        
    if tc < 0:
        raise ValueError(f'time constant (tc) must be nonnegative but got {tc}')
        
    if (isinstance(factor, float) and factor <= 0) or (isinstance(factor, int) and factor <= 0):
        raise ValueError(f'If a float, factor must be positive, but got {factor}')
    if isinstance(factor, str) and factor not in ['linear', 'boolean', 'half-wave']:
        raise ValueError(f"If a string, factor must be one of {'linear', 'boolean', 'half-wave'}, but got '{factor}'")
        
    ispow2 = (sfreq and (not(sfreq & (sfreq - 1))) )
        
    if not ispow2:
        next_pow2 = _largest_pow2_less_than(sfreq)
        x_resampled = resample(x.squeeze(), int(x.shape[0]/sfreq*next_pow2))
    else:
        x_resampled = x.squeeze().copy()
        
    shift = round(math.log2(sfreq/16384)) # octaves to shift
    
    # read in cochlear filter from file
    filedir = dirname(__file__)
    # filedir = './'
    cochba = loadmat(join(filedir, 'cochba.mat'), variable_names=['COCHBA'])
    cochba = cochba['COCHBA']
    
    L, M = cochba.shape
    L_x = x.shape[0]
    
    L_frm = round(frame_len * (2**(4+shift))) # frame length (points)
    
    if tc > 0:
        alpha = math.exp(-1/(tc * 2**(4+shift))) # decay factor
        alpha_filt = np.array([1, -alpha])
    else:
        alpha = 0 # this is now just short-term average
        
    # hair cell time constant in ms
    haircell_tc = 0.5
    beta = math.exp(-1/(haircell_tc * 2**(4+shift)))
    low_pass_filt = np.array([1, -beta])
    
    # allocate memory for output
    N = math.ceil(L_x / L_frm)
    x_resampled = np.pad(x_resampled, (N*L_frm-L_x,0))
    indexer = np.arange(L_frm-1, N*L_frm, L_frm)
    
    output = np.zeros((N, M-1))
    
    # do last channel first (highest frequency)
    
    p = int(np.real(cochba[0,M-1]))
    B = np.real(cochba[1:p+2, M-1])
    A = np.imag(cochba[1:p+2, M-1])
    
    y1 = lfilter(B, A, x_resampled).squeeze()
    y2 = _sigmoid(y1, factor)
        
    # hair cell membrane (low-pass <= 4kHz), ignored for linear ionic channels
    if factor != 'linear':
        y2 = lfilter(np.array([1.]), low_pass_filt, y2).squeeze()
    
    y2_h = y2.copy()
    y3_h = 0
    
    for ch in range(M-2, -1, -1):
        
        # cochlear filterbank
        p = int(np.real(cochba[0,ch]))
        B = np.real(cochba[1:p+2, ch])
        A = np.imag(cochba[1:p+2, ch])
        y1 = lfilter(B, A, x_resampled).squeeze()
    
        # transduction hair cells
        y2 = _sigmoid(y1, factor)
        
        # hair cell membrane (low-pass <= 4kHz), ignored for linear ionic channels
        if factor != 'linear':
            y2 = lfilter(np.array([1.]), low_pass_filt, y2).squeeze()
    
        # reduction: lateral inhibitory network
        # masked by higher (frequency) spatial response
        y3 = y2 - y2_h
        y2_h = y2.copy()
        
        # half-wave rectifier
        y4 = np.maximum(y3, 0)
        
        # temporal integration
        if alpha != 0:
            y5 = lfilter(np.array([1.]), alpha_filt, y4).squeeze()
            output[:, ch] = y5[indexer]
        else:
            if L_frm == 1:
                output[:, ch] = y4
            else:
                output[:, ch] = np.mean(np.reshape(y4, (L_frm, N)), axis=0)
        
    return output
    
    


import numpy as np

def stars(pvals):
    '''
    Convert pvalues to strings of stars for significance.
    thresholds: [inf, 0.05, 0.01, 0.001, 0.001]
    stars:    ['n.s.', '*', '**', '***', '****']
    
    Parameters
    ----------
    pvals : float or list/np.ndarray of floats
        Pvalue or pvalues to convert to stars.
    
    Returns
    -------
    stars : string or list of strings
    
    Examples
    --------
    >>> import naplib as nl
    >>> pvalues = [0.06, 0.03, 0.0008]
    >>> nl.stats.stars(pvalues)
    ['n.s.','*','****']
    '''
    thresholds = [np.inf, 0.05, 0.01, 0.001, 0.001, 0.0001]
    stars_strings = ['n.s.', '*', '**', '***', '****']
    float_flag = False
    
    stars = []
        
    if not isinstance(pvals, list) and not isinstance(pvals, np.ndarray):
        pvals = [pvals]
        float_flag = True
    for i, pval in enumerate(pvals):
        if pval < 0:
            raise ValueError(f'pvalues must not be negative but found {pval}')
        j = -1
        thresh = thresholds[0]
        while (j < len(stars_strings)-1) and pval < thresh:
            j += 1
            thresh = thresholds[j+1]
        stars.append(stars_strings[j])
    
    if float_flag:
        return stars[0]
    return stars




import copy
import contextlib
from functools import partialmethod
from contextlib import contextmanager,redirect_stderr,redirect_stdout
from os import devnull

import tqdm
from tqdm.auto import tqdm as tqdm_
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils import check_X_y
from mne.decoding import TimeDelayingRidge, ReceptiveField
from mne.decoding.time_delaying_ridge import _fit_corrs, _compute_reg_neighbors, _edge_correct
from mne.decoding.receptive_field import _reshape_for_est, _delays_to_slice, _times_to_delays, _delay_time_series, _corr_score, _r2_score

from ..out_struct import OutStruct
from ..utils import _parse_outstruct_args

from mne.cuda import _setup_cuda_fft_multiply_repeated
from mne.filter import next_fast_len
from mne.fixes import jit
from mne.parallel import check_n_jobs
from mne.utils import warn, logger


def _compute_corrs(X, y, smin, smax, n_jobs=1, fit_intercept=False,
                   edge_correction=True):
    """Compute auto- and cross-correlations.
    This class if copied from mne.decoding.time_delaying_ridge.py, but removes
    the progres bar.
    """
    if fit_intercept:
        # We could do this in the Fourier domain, too, but it should
        # be a bit cleaner numerically to do it here.
        X_offset = np.mean(X, axis=0)
        y_offset = np.mean(y, axis=0)
        if X.ndim == 3:
            X_offset = X_offset.mean(axis=0)
            y_offset = np.mean(y_offset, axis=0)
        X = X - X_offset
        y = y - y_offset
    else:
        X_offset = y_offset = 0.
    if X.ndim == 2:
        assert y.ndim == 2
        X = X[:, np.newaxis, :]
        y = y[:, np.newaxis, :]
    assert X.shape[:2] == y.shape[:2]
    len_trf = smax - smin
    len_x, n_epochs, n_ch_x = X.shape
    len_y, n_epcohs, n_ch_y = y.shape
    assert len_x == len_y

    n_fft = next_fast_len(2 * X.shape[0] - 1)

    n_jobs, cuda_dict = _setup_cuda_fft_multiply_repeated(
        n_jobs, [1.], n_fft, 'correlation calculations')

    # create our Toeplitz indexer
    ij = np.empty((len_trf, len_trf), int)
    for ii in range(len_trf):
        ij[ii, ii:] = np.arange(len_trf - ii)
        x = np.arange(n_fft - 1, n_fft - len_trf + ii, -1)
        ij[ii + 1:, ii] = x

    x_xt = np.zeros([n_ch_x * len_trf] * 2)
    x_y = np.zeros((len_trf, n_ch_x, n_ch_y), order='F')
    n = n_epochs * (n_ch_x * (n_ch_x + 1) // 2 + n_ch_x)
    logger.info('Fitting %d epochs, %d channels' % (n_epochs, n_ch_x))
    count = 0
    for ei in range(n_epochs):
        this_X = X[:, ei, :]
        # XXX maybe this is what we should parallelize over CPUs at some point
        X_fft = cuda_dict['rfft'](this_X, n=n_fft, axis=0)
        X_fft_conj = X_fft.conj()
        y_fft = cuda_dict['rfft'](y[:, ei, :], n=n_fft, axis=0)

        for ch0 in range(n_ch_x):
            for oi, ch1 in enumerate(range(ch0, n_ch_x)):
                this_result = cuda_dict['irfft'](
                    X_fft[:, ch0] * X_fft_conj[:, ch1], n=n_fft, axis=0)
                # Our autocorrelation structure is a Toeplitz matrix, but
                # it's faster to create the Toeplitz ourselves than use
                # linalg.toeplitz.
                this_result = this_result[ij]
                # However, we need to adjust for coeffs that are cut off,
                # i.e. the non-zero delays should not have the same AC value
                # as the zero-delay ones (because they actually have fewer
                # coefficients).
                #
                # These adjustments also follow a Toeplitz structure, so we
                # construct a matrix of what has been left off, compute their
                # inner products, and remove them.
                if edge_correction:
                    _edge_correct(this_result, this_X, smax, smin, ch0, ch1)

                # Store the results in our output matrix
                x_xt[ch0 * len_trf:(ch0 + 1) * len_trf,
                     ch1 * len_trf:(ch1 + 1) * len_trf] += this_result
                if ch0 != ch1:
                    x_xt[ch1 * len_trf:(ch1 + 1) * len_trf,
                         ch0 * len_trf:(ch0 + 1) * len_trf] += this_result.T
                count += 1

            # compute the crosscorrelations
            cc_temp = cuda_dict['irfft'](
                y_fft * X_fft_conj[:, slice(ch0, ch0 + 1)], n=n_fft, axis=0)
            if smin < 0 and smax >= 0:
                x_y[:-smin, ch0] += cc_temp[smin:]
                x_y[len_trf - smax:, ch0] += cc_temp[:smax]
            else:
                x_y[:, ch0] += cc_temp[smin:smax]
            count += 1

    x_y = np.reshape(x_y, (n_ch_x * len_trf, n_ch_y), order='F')
    return x_xt, x_y, n_ch_x, X_offset, y_offset


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull.
    This is used to suppress tqdm outputs from fitting, which produce
    a progress bar to sys.stderr."""
    with open(devnull, 'w') as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


class TRF(BaseEstimator):
    '''
    Allows the fitting of temporal receptive field (TRF) models
    to one or more targets at a time using cross-validation. These can be
    encoding models (stimulus-to-brain) or decoding (brain-to-stimulus) models.
    Internally, this fits several mne.decoding.ReceptiveField models, one for each
    target variable.

    Please see the :ref:`TRF example notebooks <STRF examples>` for more detailed
    tutorials which show how to train, test, and inspect TRF and STRF models.
    
    Parameters
    ----------
    tmin : float
        The starting lag (inclusive), in seconds (or samples if ``sfreq`` == 1).
    tmax : float
        The ending lag (noninclusive), in seconds (or samples if ``sfreq`` == 1).
        Must be > tmin.
    sfreq : int
        The sampling frequency used to convert times into samples.
    reg_type : str, default='ridge'
        Regularization type. Can be "ridge" (default) or "laplacian".
    alpha : float | list or array-like of floats, default=np.logspace(2, 9, 8).
        Regularization strength. If a list or array-like of values, then the best
        one will be fit with cross-validation. If a list is given, alpha is
        optimized for each target variable individually.
    xval_test_portion : float, default=0.25
        If multiple alpha values are given, cross-validataion is performed to
        choose the best, using this portion as withheld test data on each
        cross-validation loop.
    fit_intercept : bool | None, default=False
        If True, the sample mean is removed before fitting.
    scoring : str, default='corrcoef'
        Defines how predictions will be scored. Currently must be one of
        'r2' (coefficient of determination) or 'corrcoef' (the correlation
        coefficient).
    n_jobs : int | str
        Number of jobs to run in parallel. Can be 'cuda' if CuPy
        is installed properly and ``estimator is None``.
    verbose : int, default=1
        Level of printing output desired.
        0 prints nothing, 1 prints only a single tqdm progress bar for
        the fitting over the n-outputs in .fit(), and 2 prints information about
        cross-validation, such as the alpha value chosen and the corresponding scores,
        during the fitting procedure for each output.
        

    Notes
    -----
    For a causal system, the encoding model would have significant
    non-zero values only at positive lags. In other words, lags point
    backwards in time relative to the input, so positive lags correspond
    to previous time samples, while negative lags correspond to future
    time samples. In most cases, an encoding model should use tmin=0
    and tmax>0, while a decoding model should use tmin<0 and tmax=0.
    
    '''
    def __init__(self,
                 tmin,
                 tmax,
                 sfreq,
                 reg_type='ridge',
                 alpha=None,
                 xval_test_portion=0.25,
                 fit_intercept=False,
                 scoring='corrcoef',
                 n_jobs=1,
                 verbose=1):
        
        if tmin >= tmax:
            raise ValueError(f'tmin must be less than tmax, but got {tmin} and {tmax}')
        
        if alpha is None:
            alpha = [round(x, 2) for x in np.logspace(2, 9, 8)]
        
        self.sfreq = float(sfreq)
        self.tmin = tmin
        self.tmax = tmax
        self.reg_type = reg_type
        self.alpha = np.array([alpha]) if isinstance(alpha, float) or isinstance(alpha, int) else alpha
        self.xval_test_portion = xval_test_portion
        self.fit_intercept = fit_intercept
        self.scoring = scoring
        self.n_jobs = n_jobs
        self.verbose = verbose
        
        if len(self.alpha) > 1 and self.xval_test_portion > 0.5:
            raise ValueError(f'xval_test_portion must be no more than 0.5 if multiple alphas were given,'+
                            f' so that cross validation can occur correctly, but got {xval_test_portion}')
        
    @property
    def _smin(self):
        return int(round(self.tmin * self.sfreq))

    @property
    def _smax(self):
        return int(round(self.tmax * self.sfreq)) + 1
        
    def _delay_and_reshape(self, X, y=None):
        """Delay and reshape the variables.
        X and y should be arrays.
        """
        if not isinstance(self.estimator_, TimeDelayingRidge):
            # X is now shape (n_times, n_epochs, n_feats, n_delays)
            X = _delay_time_series(X, self.tmin, self.tmax, self.sfreq,
                                   fill_mean=self.fit_intercept)
            X = _reshape_for_est(X)
            # Concat times + epochs
            if y is not None:
                y = y.reshape(-1, y.shape[-1], order='F')
        return X, y

    def fit(self, outstruct=None, X='aud', y='resp'):
        '''
        Fit a multi-output model to the data in X and y, which contain multiple trials.
        
        Parameters
        ----------
        outstruct : naplib.OutStruct object, optional
            OutStruct containing data to be normalized in one of the field.
            If not given, must give the X and y data directly as the ``X``
            and ``y`` arguments. 
        X : str | list of np.ndarrays or a multidimensional np.ndarray
            Data to be used as predictor in the regression. Once arranged,
            should be of shape (time, num_features).
            If a string, it must specify one of the fields of the outstruct
            provided in the first argument. If a multidimensional array, first dimension
            indicates the trial/instances which will be concatenated over to compute
            normalization statistics.
        y : str | list of np.ndarrays or a multidimensional np.ndarray
            Data to be used as target(s) in the regression. Once arranged,
            should be of shape (time, num_targets[, num_features_y]).
            If a string, it must specify one of the fields of the outstruct
            provided in the first argument. If a multidimensional array, first dimension
            indicates the trial/instances which will be concatenated over to compute
            normalization statistics.
        
                
        Returns
        -------
        self : returns an instance of self
        
        '''
        
        X_, y_ = _parse_outstruct_args(outstruct, copy.deepcopy(X), copy.deepcopy(y))
        
        X_ = np.concatenate(X_, axis=0)

        if y_[0].ndim == 1:
            y_ = [yy.reshape(-1,1) for yy in y_]
            
        # smooth beginning and ending of each target
        hamming = np.hamming(2*int((self.tmax-self.tmin) * self.sfreq))
        hamming_start = hamming[:len(hamming)//2].reshape(-1,1)
        hamming_end = hamming[-len(hamming)//2:].reshape(-1,1)
        if y_[0].ndim == 3:
            hamming_start = np.concatenate([hamming_start.reshape(-1,1,1) for _ in range(y_[0].shape[-1])], axis=-1)
            hamming_end = np.concatenate([hamming_end.reshape(-1,1,1) for _ in range(y_[0].shape[-1])], axis=-1)
        for i in range(len(y_)):
            y_[i][:len(hamming_start)] = np.multiply(y_[i][:len(hamming_start)], hamming_start)
            y_[i][-len(hamming_end):] = np.multiply(y_[i][-len(hamming_end):], hamming_end)

            
        y_ = np.concatenate(y_, axis=0)
        
        
        self.ndim_y_ = y_.ndim # either 2 or 3, depending on whether each samples of y has multiple output features
        if self.ndim_y_ == 3:
            X_ = X_[:, np.newaxis, :]
        # X and y should now both be arrays of shape (time, num_features) and (time, num_targets), respectively
        
        self.models_ = []
        
        # Initialize delays
        self.delays_ = _times_to_delays(self.tmin, self.tmax, self.sfreq)
        
        # Define the slice that we should use in the middle
        self.valid_samples_ = _delays_to_slice(self.delays_)
        
        # for each target variable, fit a TRF model with cross-validation
        if self.verbose >= 1:
            disable_tqdm = False
        else:
            disable_tqdm = True

        for target_idx in tqdm_(range(y_.shape[1]), disable=disable_tqdm):
            
            if self.verbose >= 2:
                print(f'Fitting model for output variable {target_idx}...')
            
            y_thistarget = y_[:,target_idx]
            if self.ndim_y_ == 2:
                y_thistarget = y_thistarget[:,np.newaxis]
            else:
                y_thistarget = y_thistarget[:,np.newaxis,:]
            
            if len(self.alpha) > 1:
                if self.verbose >= 2:
                    print(f'Running cross-validation on alphas...')
                
                scores = []
                
                # for each alpha, do cross-validation to get score
                
                X_generator = _xval_time_splits(X_, self.xval_test_portion)
                y_generator = _xval_time_splits(y_thistarget, self.xval_test_portion)
                for (X_train, X_test), (y_train, y_test) in zip(X_generator, y_generator):
                    scores_thissplit = []
                                        
                    rf = self._get_single_RF_model(verbose=False)
                    rf._y_dim = self.ndim_y_
                    self.estimator_ = rf.estimator
                    
                    # take the expensive computation of covariance outside the alpha loop,
                    # and only do it once for each train-test split
                    
                    rf.delays_ = self.delays_
                    rf.valid_samples_ = self.valid_samples_
                    
                    # Create input features
                    X_train_delayed, y_train_delayed = self._delay_and_reshape(X_train, y_train)
                    
                    # use this context manager to suppress tqdm output
                    with suppress_stdout_stderr():
                        cov_, x_y_, n_ch_x, X_offset, y_offset = _compute_corrs(
                            X_train_delayed, y_train_delayed, self._smin, self._smax, self.n_jobs,
                            self.fit_intercept, edge_correction=True)
                    rf.cov_ = cov_
                    rf.estimator.cov_ = cov_
                    rf.estimator_ = rf.estimator
                    
                    for ii, alpha_ in enumerate(self.alpha):
                        
                        coef_ = _fit_corrs(cov_, x_y_, n_ch_x,
                                self.reg_type, alpha_, n_ch_x)
                        coef_ = coef_[:,:-1]
                        
                        rf.coef_ = coef_
                        rf.estimator_.coef_ = coef_

                        if self.fit_intercept:
                            rf.estimator_.intercept_ = y_offset - np.dot(X_offset, self.coef_.sum(-1).T)
                        else:
                            rf.estimator_.intercept_ = 0.

                        # Now make predictions about the model output, given input stimuli.
                        y_pred_test = rf.predict(X_test)
                        scores_thissplit.append(self._score_predictions([y_pred_test], [y_test], [1.]))

                    scores.append(scores_thissplit)
                scores = np.array(scores).squeeze()
                mean_scores = scores.mean(0) # average score for each alpha over all xval folds
                
                # choose best alpha
                ix_best_alpha_lap = np.argmax(mean_scores)
                best_alpha_thistarget = self.alpha[ix_best_alpha_lap]
                _print_alpha_xvals(self.alpha, mean_scores, verbose=self.verbose)
                if self.verbose >= 2:
                    print('chose alpha={:.2e}\n'.format(best_alpha_thistarget))
            else:
                best_alpha_thistarget = self.alpha[0]
                
            rf = self._get_single_RF_model(alpha_val=best_alpha_thistarget)
            rf._y_dim = self.ndim_y_
            rf.delays_ = self.delays_
            rf.valid_samples_ = self.valid_samples_
            self.estimator_ = rf.estimator
            
            X_train_delayed, y_train_delayed = self._delay_and_reshape(X_, y_thistarget)
            # use this context manager to suppress tqdm output
            with suppress_stdout_stderr():
                cov_, x_y_, n_ch_x, X_offset, y_offset = _compute_corrs(
                            X_train_delayed, y_train_delayed, self._smin, self._smax, self.n_jobs,
                            self.fit_intercept, edge_correction=True)
            rf.cov_ = cov_
            rf.estimator.cov_ = cov_
            rf.estimator_ = rf.estimator
            coef_ = _fit_corrs(cov_, x_y_, n_ch_x,
                            self.reg_type, best_alpha_thistarget, n_ch_x)
            coef_ = coef_[:,:-1]

            rf.coef_ = coef_
            rf.estimator_.coef_ = coef_

            if self.fit_intercept:
                rf.estimator_.intercept_ = y_offset - np.dot(X_offset, self.coef_.sum(-1).T)
            else:
                rf.estimator_.intercept_ = 0.
            # remove last lag of weights because it has significant edge effect
            rf.estimator_.coef_ = rf.coef_[:,:,:-1]
            rf.coef_ = rf.coef_[:,:,:-1] 
            self.models_.append(rf)
            
        return self
    
    @property
    def coef_(self):
        if not hasattr(self, 'ndim_y_'):
            raise ValueError(f'Must call fit() first before accessing coef_ attribute.')
        coefs_ = []
        for mdl in self.models_:
            if self.ndim_y_ == 2:
                coefs_.append(mdl.coef_)
            else:
                coefs_.append(mdl.coef_[np.newaxis,:,:,:])
                
        return np.concatenate(coefs_, axis=0) # shape (num_outputs, freqs, lags)
            
    def _get_single_RF_model(self, alpha_val=0, verbose=0):
        estimator = TimeDelayingRidge(tmin=self.tmin, tmax=self.tmax, sfreq=self.sfreq, reg_type=self.reg_type,
                                      fit_intercept=self.fit_intercept, alpha=alpha_val)
        rf = ReceptiveField(tmin=self.tmin, tmax=self.tmax, sfreq=self.sfreq, estimator=estimator,
                            fit_intercept=self.fit_intercept, scoring=self.scoring, n_jobs=self.n_jobs, verbose=verbose)
        return rf
    
    def predict(self, outstruct=None, X='aud'):
        '''
        Parameters
        ----------
        outstruct : naplib.OutStruct object, optional
            OutStruct containing data to be normalized in one of the field.
            If not given, must give the X and y data directly as the ``X``
            and ``y`` arguments. 
        X : str | list of np.ndarrays or a multidimensional np.ndarray
            Data to be used as predictor in the regression. Once arranged,
            should be of shape (time, num_features).
            If a string, it must specify one of the fields of the outstruct
            provided in the first argument. If a multidimensional array, first dimension
            indicates the trial/instances which will be concatenated over to compute
            normalization statistics.
        
        Returns
        -------
        y_pred : np.ndarray or list of np.ndarrays, each with shape (time, num_targets[, num_features_y])
            Predicted target for each trial in X.
        '''
        
        if not hasattr(self, 'models_'):
            raise ValueError(f'Must call .fit() before can call .predict()')
        
        X = _parse_outstruct_args(outstruct, X)
        
        y_pred = [[] for _ in X]
        for ii, x_trial in enumerate(X):
            for mdl in self.models_:
                tmp_pred = mdl.predict(x_trial)
                
                if self.ndim_y_ == 2:
                    tmp_pred = tmp_pred.reshape(-1,1)
                else:
                    tmp_pred = tmp_pred.reshape(tmp_pred.shape[0],1,-1)
                y_pred[ii].append(tmp_pred)
            y_pred[ii] = np.concatenate(y_pred[ii], axis=1)
        
        return y_pred
    
    def score(self, outstruct=None, X='aud', y='resp', weight_trial_duration=True):
        '''
        Get scores from predictions of the model.
        
        Parameters
        ----------
        outstruct : naplib.OutStruct object, optional
            OutStruct containing data to be normalized in one of the field.
            If not given, must give the X and y data directly as the ``X``
            and ``y`` arguments. 
        X : str | list of np.ndarrays or a multidimensional np.ndarray
            Data to be used as predictor in the regression. Once arranged,
            should be of shape (time, num_features).
            If a string, it must specify one of the fields of the outstruct
            provided in the first argument. If a multidimensional array, first dimension
            indicates the trial/instances which will be concatenated over to compute
            normalization statistics.
        y : str | list of np.ndarrays or a multidimensional np.ndarray
            Data to be used as target(s) in the regression. Once arranged,
            should be of shape (time, num_targets[, num_features_y]).
            If a string, it must specify one of the fields of the outstruct
            provided in the first argument. If a multidimensional array, first dimension
            indicates the trial/instances which will be concatenated over to compute
            normalization statistics.
        weight_trial_duration : bool, default=True
            Scores are computed for each trial and then aggregated by
            averaging across trials in X. If set to True (default), this
            averaging is weighted by the duration of each trial.

        Returns
        -------
        scores : np.array of float, shape (n_outputs,)
            The scores estimated by the model for each output (e.g. mean
            R2 of ``predict(X)``).
        '''
        
        if not hasattr(self, 'models_'):
            raise ValueError(f'Must call .fit() before can call .score()')
        
        X_, y_ = _parse_outstruct_args(outstruct, copy.deepcopy(X), copy.deepcopy(y))        

        y_preds = []
        for x, yy in zip(X_, y_):
            if yy.ndim != self.ndim_y_:
                raise ValueError(f'True y values are incorrect shape. Should be {self.ndim_y_} dimensions, but found a trial with {yy.ndim}')
            y_preds.append(self.predict(X=[x])[0])
            
            
        if weight_trial_duration:
            weights = [x.shape[0] for x in X_]
            weights = [w / float(sum(weights)) for w in weights]
        else:
            weights = [1. for _ in X_]
            
        return self._score_predictions(y_preds, y_, weights)
    

    def _score_predictions(self, y_preds_list, y_true_list, weights):
        
        scorer = _SCORERS[self.scoring]
        
        # for each trial, compute scores
        y_scores = []
        for y_pred, y_true in zip(y_preds_list, y_true_list):
            assert y_true.ndim == self.ndim_y_
            if self.ndim_y_ == 3:
                y_pred = y_pred.transpose(0,2,1).reshape(-1, y_pred.shape[1]) # move output channel dimension to end, then flatten all other features over time
                y_true_reshaped = y_true.transpose(0,2,1).reshape(-1, y_true.shape[1])
            else:
                y_true_reshaped = y_true
            assert y_pred.shape[1] == y_true.shape[1]
            y_scores.append(scorer(y_pred, y_true_reshaped, multioutput='raw_values')[:,np.newaxis])
        y_scores = np.concatenate(y_scores, axis=1)
        
        return np.average(y_scores, axis=1, weights=weights)
            
def _print_alpha_xvals(alphas, alpha_scores, verbose=1):
    
    if verbose >= 2:
        print('Cross-validation scores:')
        for alpha_i, score_i in zip(alphas, alpha_scores):
            print('alpha={:.2e} : {:.4f}'.format(alpha_i, score_i))
                
def _xval_time_splits(data, test_portion):
    '''
    data : np.array of shape (time, num_features)
    test_portion : float between 0 and 1
    '''
    win_size = np.floor(data.shape[0] * test_portion).astype('int')
    
    num_chunks = int(((data.shape[0]-win_size)/win_size)+1)
    
    idx = np.arange(num_chunks*win_size)
    
    for i in range(0,num_chunks*win_size,win_size):
        start_idx = int(idx[i])
        stop_idx = int(idx[i] + win_size)
        train_data = np.delete(data, slice(start_idx, stop_idx), axis=0)
        test_data = data[start_idx:stop_idx]
        yield train_data, test_data

_SCORERS = {'r2': _r2_score, 'corrcoef': _corr_score}

